from __future__ import annotations

import ctypes
import inspect
import re
import threading
import time
import warnings
from collections import defaultdict
from enum import Enum
from functools import wraps
from queue import Queue
from threading import Thread
from types import FrameType
from typing import (TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple,
                    Union)

import torch

from ..util import applyn
from .batching import Batcher
from .tracing.util import wrap_exception, get_non_nnsight_frame, push_variables

if TYPE_CHECKING:
    from .tracing.tracer import Cache, InterleavingTracer, Tracer


class Events(Enum):
    """Enum for different types of events in the interleaving process."""

    VALUE = "value"  # Request for a value
    SWAP = "swap"  # Request for a swap
    END = "end"  # Signal to end the execution
    EXCEPTION = "exception"  # Signal that an exception occurred
    SKIP = "skip"  # Signal that an operation should be skipped
    REGISTER = "register"  # Signal that a child mediator should be registered
    BARRIER = "barrier"  # Signal that a barrier should be set


class Cancelation(Exception):
    """Exception raised when a request is canceled."""

    pass


class EarlyStopException(Exception):
    """
    Exception raised to stop the execution of the model.
    """

    pass


class SkipException(Exception):
    """
    Exception raised to skip the execution of the model.
    """

    def __init__(self, value: Any):
        self.value = value


class Interleaver:
    """
    Manages the interleaving of model execution and interventions.

    This class coordinates the flow between the model's forward pass and
    user-defined intervention functions, allowing for inspection and
    modification of intermediate values.
    """
    
    def __init__(
        self,
        invokers: List[Mediator] = None,
        tracer: InterleavingTracer = None,
        batcher: Batcher = None,
        user_cache: Optional[Cache] = None,
    ):
        self.initialize(invokers, tracer, batcher, user_cache)


    def initialize(
        self,
        invokers: List[Mediator],
        tracer: InterleavingTracer,
        batcher: Batcher = None,
        user_cache: Optional[Cache] = None,
    ):

        self.invokers = invokers
        self.tracer = tracer
        self.batcher = batcher if batcher is not None else Batcher()
        self.user_cache = user_cache

        self.mediators: Dict[str, Mediator] = {}
        self.iteration_tracker = defaultdict(int)
        self.default_all = None
        
    def cancel(self):
        """Cancel all intervention threads."""
     
        for mediator in list(self.mediators.values()):
            mediator.cancel()

        self.mediators = None
        self.tracer = None
        self.batcher = None
        self.user_cache = None
        self.invokers = None


    def iterate(self, provider: str):

        iteration = self.iteration_tracker[provider]

        self.iteration_tracker[provider] += 1

        return f"{provider}.i{iteration}"

    def wrap_module(self, module: torch.nn.Module):

        skip = None

        forward = module.forward

        @wraps(module.forward)
        def skippable_forward(*args, **kwargs):

            nonlocal skip

            if skip is None or not self.interleaving:
                try:
                    return forward(*args, **kwargs)
                finally:
                    skip = None

            return skip

        module.forward = skippable_forward

        @torch._dynamo.disable
        def input_hook(module: torch.nn.Module, args, kwargs):

            if not self.interleaving:
                return args, kwargs

            provider = module.__path__

            nonlocal skip

            try:
                inputs = self.handle(f"{provider}.input", (args, kwargs), iterate=True)
            except SkipException as e:
                skip = e.value
            else:
                args, kwargs = inputs

            return args, kwargs

        module.register_forward_pre_hook(input_hook, with_kwargs=True, prepend=True)

        @torch._dynamo.disable
        def output_hook(module: torch.nn.Module, _, output: Any):

            if not self.interleaving:
                return output

            provider = module.__path__

            nonlocal skip

            if skip is not None:

                output = skip

                skip = None

            output = self.handle(f"{provider}.output", output, iterate=True)

            return output

        module.register_forward_hook(output_hook, prepend=True)

    def wrap_operation(self, fn: Callable, name: str, bound_obj: Optional[Any] = None):
        """
        Wrap an operation to intercept inputs and outputs for intervention.

        Args:
            fn: The function to wrap
            name: The name of the operation
            bound_obj: The object fn is bound to if it is a method

        Returns:
            A wrapped version of the function
        """

        @wraps(fn)
        def inner(*args, **kwargs):

            nonlocal fn

            fn = self.handle(f"{name}.fn", fn)

            args, kwargs = self.handle(f"{name}.input", (args, kwargs))

            if not inspect.ismethod(fn) and bound_obj is not None:
                value = fn(bound_obj, *args, **kwargs)
            else:
                value = fn(*args, **kwargs)

            value = self.handle(f"{name}.output", value)

            return value

        return inner


    @property
    def interleaving(self):
        return getattr(self, "_interleaving", False)

    def __enter__(self):
        
        self._interleaving = True
        
        try:
            for invoker in self.invokers:
                invoker.start(self)

            try:
                self.handle()
            except EarlyStopException:
                pass
        except:
            self._interleaving = False
            raise

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self._interleaving = False
        
        
        # If execution was stopped early, ignore and do nothing
        if exc_type is not None and issubclass(exc_type, EarlyStopException):
            return True
        
    def check_dangling_mediators(self):
        
        # If any mediators are still waiting for their values for their events, they probably called an Envoy out of order
        # Or their Envoy was not called.
        for mediator in self.mediators.values():

            if mediator.child is not None:
                mediator = mediator.child

            if not mediator.event_queue.empty():
                requested_event, requester = mediator.event_queue.get()

                if isinstance(requester, tuple):
                    requester = requester[0]

                mediator.respond(
                    ValueError(
                        f"Execution complete but `{requester}` was not provided. Did you call an Envoy out of order? Investigate why this module was not called?"
                    )
                )
                mediator.wait()

                if mediator.name.startswith("Iterator"):
                    try:
                        mediator.handle()
                    except ValueError as e:
                        msg = f"Execution complete but `{requester}` was not provided. This was in an Iterator at iteration {mediator.iteration} so likely this iteration did not happen. If you were using `.iter[:]`, this is likely not an error."
                        warnings.warn(msg)
                else:
                    mediator.handle()
                    
    def check_cache_full(self):
        """
        Print a warning if a module to be cached was missed.
        """
        for invoker in self.invokers:
            for cache in invoker.user_cache:
                if cache.modules:
                    if cache.include_inputs and cache.include_output:
                        for module in cache.modules:
                            if (
                                module not in cache.cache
                                or cache.cache[module].inputs is None
                            ):
                                print(
                                    "\033[33m"
                                    + "NNsight Warning: A module to be cached was missed! Consider defining the Cache before the module is called."
                                    + "\033[0m"
                                )
                                return
                    else:
                        if any(module not in cache.cache for module in cache.modules):
                            print(
                                "\033[33m"
                                + "NNsight Warning: A module to be cached was missed! Consider defining the Cache before the module is called."
                                + "\033[0m"
                            )
                            return

    ### Provider Methods ###
    
    def handle(
        self,
        provider: Optional[Any] = None,
        value: Optional[Any] = None,
        iterate: bool = False,
    ):
        """
        Handle a provider's value, allowing mediators to consume and modify it.

        Args:
            provider: The identifier of the provider
            value: The value being provided

        Returns:
            The original or modified value
        """

        if iterate:
            provider = self.iterate(provider)

        old = self.batcher.current_value

        self.batcher.current_value = value

        skip_count = 0
        skip_values = []

        for mediator in self.invokers:

            try:
                mediator.handle(provider)
            except SkipException as e:
                skip_count += 1
                skip_values.append(e.value)

        if skip_count == len(self.invokers) and self.invokers:

            def _swap(*args):
                return torch.cat(args, dim=0)

            skip_value = applyn(skip_values, _swap, torch.Tensor)
            raise SkipException(skip_value)
        elif skip_count > 0 and skip_count < len(self.invokers):
            raise ValueError(
                f"A module skip must be applied to all the invokers defined in the tracer!"
            )

        value = self.batcher.current_value

        self.batcher.current_value = old

        if (
            self.user_cache is not None
            and len(self.user_cache) > 0
            and provider is not None
        ):
            for cache in self.user_cache:
                cache.add(provider, value)

        return value

    ### Requester Methods ###

    @property
    def current(self) -> Mediator:
        """Get the current mediator."""
        return self.mediators[threading.current_thread().name]

    ### Serialization ###

    def __deepcopy__(self, memo):

        return self


class Mediator:
    """
    Mediates between the model execution and intervention functions.

    This class handles the communication between the model's forward pass and
    user-defined intervention functions, allowing for inspection and
    modification of intermediate values.
    """

    class OutOfOrderError(Exception):
        """
        Exception raised when interventions are defined out of order.
        """

        pass

    def __init__(
        self,
        intervention: Callable,
        info: "Tracer.Info",
        name: Optional[str] = None,
        batch_group: Optional[int] = 0,
        stop: Optional[int] = None,
    ) -> None:
        """
        Initialize a Mediator with an intervention function.

        Args:
            intervention: The intervention function
            info: Information about the tracing context
            name: Optional name for the mediator
            stop: Optional number of times to execute this mediator
        """
        self.intervention = intervention
        self.name = name if name else f"Mediator{id(self)}"
        self.info = info
        self.batch_group = batch_group
        self.event_queue = Queue()
        self.response_queue = Queue()
        
        self.child: Mediator = None

        self.thread = None
        self.interleaver = None
        self.history = set()
        self.user_cache: List["Cache"] = list()
        self.iteration = 0
        self.all_stop: Optional[int] = stop

        self.args = list()

    @property
    def alive(self):
        return self.thread is not None and self.thread.is_alive()

    def start(self, interleaver: Interleaver):
        """
        Start the mediator's intervention thread.

        Args:
            interleaver: The interleaver managing this mediator
        """
        self.interleaver = interleaver

        self.interleaver.mediators[self.name] = self

        if not self.alive:

            self.thread = Thread(
                target=self.intervention,
                args=(self, self.info, *self.args),
                daemon=True,
                name=self.name,
            )
            self.thread.start()

            self.wait()

    ### Provider Methods ###

    def wait(self):
        """Wait for the next event to be set in the event queue."""
        while self.event_queue.empty() and self.alive:
            # Keep checking until there's an event in the queue
            time.sleep(0.001)  # Small sleep to prevent CPU spinning

    def cancel(self):
        """Cancel the intervention thread and clear caches."""
        # TODO custom canceled error

        self.interleaver.mediators.pop(self.name)

        self.history.clear()

        self.thread = None

        if self.alive:
            # TODO: cancel inactive threads at the end of the model's execution
            self.response_queue.put(Cancelation())

    def handle(self, provider: Optional[Any] = None):
        """
        Handle events in the event queue and process provider values.

        Args:
            provider: The identifier of the provider

        Returns:
            The original or modified value
        """
        
        if self.child is not None:
            self.child.handle(provider)
            
            if not self.child.alive:
                self.child = None
                self.respond()
            else:
                return
            
        process = not self.event_queue.empty()

        event = None
        
        while process:
            
            value = self.interleaver.batcher.current_value

            event, data = self.event_queue.get()

            if event == Events.VALUE:
                process = self.handle_value_event(data, provider, value)
            elif event == Events.SWAP:
                process = self.handle_swap_event(provider, *data)
            elif event == Events.EXCEPTION:
                process = self.handle_exception_event(data)
            elif event == Events.SKIP:
                try:
                    process = self.handle_skip_event(provider, *data)
                except SkipException as e:
                    if len(self.user_cache) > 0:
                        for cache in self.user_cache:
                            cache.add(provider, e.value)
                    raise e
            elif event == Events.REGISTER:
                process = self.handle_register_event(provider, data)
            elif event == Events.BARRIER:
                process = self.handle_barrier_event(provider, data)
            elif event == Events.END:
                process = False

        if event == Events.END:
            self.handle_end_event()

        # TODO maybe move this to the interleaver to cache the pre-iteration provider
        if len(self.user_cache) > 0 and provider is not None:

            for cache in self.user_cache:
                cache.add(
                    provider,
                    self.interleaver.batcher.narrow(
                        self.batch_group, self.interleaver.batcher.current_value
                    ),
                )
                
    def handle_register_event(self, provider: Any, child:Mediator) -> bool:
                
        self.child = child
        child.start(self.interleaver)
        child.handle(provider)
        
        return False

    def handle_barrier_event(self, provider: Any, participants: Set[str]):
        """
        Handle a barrier event by setting a barrier.
        """
        
        if participants is not None:
        
            for mediator in self.interleaver.invokers:
                
                while mediator.child is not None:
                    mediator = mediator.child
                    
                if mediator.name in participants:
               
                    mediator.respond()
        
                    mediator.handle(provider)

                    
                    
                            


    def handle_end_event(self):
        """
        Handle an end event by stopping the mediator.
        """
        self.cancel()

    def handle_value_event(self, requester: Any, provider: Any, value: Any):
        """
        Handle a value event by providing the requested value or recording a missed provider.

        Args:
            requester: The identifier of the requester
            provider: The identifier of the provider
            value: The value being provided

        Returns:
            Boolean indicating whether to continue processing events
        """

        if provider == requester:

            value = self.interleaver.batcher.narrow(self.batch_group, value)

            self.respond(value)
        else:
            if requester in self.history:
                # TODO needs tests
                self.respond(
                    Mediator.OutOfOrderError(
                        f"Value was missed for {requester}. Did you call an Envoy out of order?"
                    )
                )
            else:
                self.history.add(provider)
                self.event_queue.put((Events.VALUE, requester))

                return False

        return True

    def handle_swap_event(self, provider: Any, requester: Any, swap_value: Any):
        """
        Handle a swap event by swapping the value if the provider matches the requester.

        Args:
            requester: The identifier of the requester
            provider: The identifier of the provider
            swap_value: The value to swap in

        Returns:
            Boolean indicating whether to continue processing events
        """
        if provider == requester:

            self.interleaver.batcher.swap(self.batch_group, swap_value)

            self.respond()

            return True

        else:
            if requester in self.history:
                # TODO needs tests
                self.respond(
                    ValueError(
                        f"Setting {requester} is out of scope for scope {provider}. Did you call an Envoy out of order?"
                    )
                )
            else:
                self.history.add(provider)
                self.event_queue.put((Events.SWAP, (requester, swap_value)))

                return False

        return True

    def handle_exception_event(self, exception: Exception):
        """
        Handle an exception event by raising the exception.

        Args:
            exception: The exception to raise

        Returns:
            Boolean indicating whether to continue processing events
        """

        if not isinstance(exception, Cancelation):

            exception = wrap_exception(exception, self.info)

            raise exception

        return False

    def handle_skip_event(self, provider: Any, requester: Any, value: Any):

        if provider == requester:

            self.respond()

            self.history.add(provider)

            raise SkipException(value)

        else:
            if requester in self.history:
                self.respond(
                    Mediator.OutOfOrderError(
                        f"Value was missed for {requester}. Did you call an Envoy out of order?"
                    )
                )

                return True
            else:
                self.history.add(provider)
                self.event_queue.put((Events.SKIP, (requester, value)))

                return False

    def respond(self, value: Optional[Any] = None):
        """
        Set the value for a pending value request.

        Args:
            value: The value to provide
        """

        self.response_queue.put(value)

        self.wait()

    ### Requester Methods ###

    @property
    def frame(self) -> FrameType:
        """
        Get the frame of the intervention function.

        Returns:
            The frame of the intervention function
        """
        
       

        frame = get_non_nnsight_frame()

        return frame

    def iterate(self, requester: Any):

        return f"{requester}.i{self.iteration}"

    def push(self):
        """Push local variables to the interleaver state."""
        
        state = {k: v for k, v in self.frame.f_locals.items() if not k.startswith("__nnsight")}
         # this does not handle the case of a fn thats called in an invoker. this will push vars directly to where the invoke was called not the fn. really we need to grad the f_back of the <nnsight> frame. If its in threading.py, then we use info.frame
        push_variables(self.info.frame, state)

    def pull(self):
        """Pull variables from the interleaver state to the frame globals."""

        state = {k: v for k, v in self.info.frame.f_locals.items() if not k.startswith("__nnsight") and k not in self.frame.f_locals}

        push_variables(self.frame, state)

    def send(self, event: Events, requester: Any):
        """
        Send an event to the event queue and wait for a response.

        Args:
            event: The event to send
            requester: The identifier of the requester

        Returns:
            The response from the provider
        """
        self.push()

        self.event_queue.put((event, requester))

        response = self.response_queue.get()

        self.pull()

        if isinstance(response, Exception):
            raise response

        return response

    def request(self, requester: Any):
        """
        Request a value from a specific provider.

        Args:
            requester: The identifier of the provider to request a value from

        Returns:
            The requested value
        """

        value = self.send(Events.VALUE, requester)

        return value

    def swap(self, requester: Any, value: Any):
        """
        Set a value to swap during execution.

        Args:
            requester: The identifier of the requester
            value: The value to swap in
        """

        self.send(Events.SWAP, (requester, value))

    def iter(self, mediator: Mediator, iteration: Union[int, slice]):
        """
        Iterate a mediator a specified number of times.

        Args:
            mediator: The mediator to iterate
            iteration: The number of iterations
        """

        def do_iteration(iter: int):

            mediator.iteration = iter
            mediator.args = list([mediator.iteration])
            
            self.send(Events.REGISTER, mediator)

        if isinstance(iteration, slice):

            i = iteration.start if iteration.start is not None else self.iteration

            stop = iteration.stop

            while True:

                do_iteration(i)

                if stop is None:
                    if self.all_stop is not None:
                        stop = self.all_stop
                
                    elif self.interleaver.default_all is not None:
                        stop = self.interleaver.default_all

                i += 1

                if stop is not None and i >= stop:
                    break

        elif isinstance(iteration, list):

            iteration.sort()

            for i in iteration:
                do_iteration(i)

        elif isinstance(iteration, int):

            do_iteration(iteration)

    def stop(self):
        """Stop the execution of the model by raising an EarlyStopException."""

        self.push()

        raise EarlyStopException()

    def skip(self, requester: Any, value: Any):

        self.send(Events.SKIP, (requester, value))

    def end(self):
        """Signal that execution should continue without further intervention."""

        self.push()

        self.event_queue.put((Events.END, None))

    def exception(self, exception: Exception):
        """
        Signal that an exception occurred during intervention.

        Args:
            exception: The exception that occurred
        """
        self.event_queue.put((Events.EXCEPTION, exception))

    def set_user_cache(self, cache: "Cache"):
        """
        Set the user cache for this mediator.

        Args:
            cache: The cache to set
        """

        self.user_cache.append(cache)

    ### Serialization ###

    def __getstate__(self):
        """Get the state of the mediator for serialization."""

        return {
            "name": self.name,
            "info": self.info,
            "batch_group": self.batch_group,
            "intervention": self.intervention,
        }

    def __setstate__(self, state):
        """Set the state of the mediator for deserialization."""
        self.name = state["name"]
        self.info = state["info"]
        self.batch_group = state["batch_group"]
        self.intervention = state["intervention"]

        self.event_queue = Queue()
        self.response_queue = Queue()

        self.thread = None
        self.interleaver = None
        self.child: Mediator = None
        self.history = set()
        self.user_cache: "Cache" = list()
        self.iteration = 0
        self.args = list()



import copy
from typing import Callable, Dict, List, Optional, Tuple

import torch

from vllm.model_executor.sampling_metadata import (
    SamplingMetadata,
    SamplingMetadataCache,
    _prepare_seq_groups,
)
from vllm.sampling_params import SamplingParams
from vllm.utils import async_tensor_h2d
from vllm.v1.sample.metadata import SamplingMetadata
from ...intervention.interleaver import Mediator
from ...intervention.tracing.tracer import Tracer


class NNsightSamplingParams(SamplingParams):

    mediator: Optional[Mediator] = None
    nns_batch_groups: Optional[List[Tuple[int, int]]] = None
    invoker_group: Optional[int] = None
    
    is_default_param: bool = True

    def clone(self) -> "SamplingParams":
        """Deep copy excluding LogitsProcessor objects.

        LogitsProcessor objects are excluded because they may contain an
        arbitrary, nontrivial amount of data.
        See https://github.com/vllm-project/vllm/issues/3087
        """

        memo = {}

        if self.logits_processors is not None:
            for lp in self.logits_processors:
                memo[id(lp)] = lp

        if self.intervention_graph is not None:
            memo[id(self.intervention_graph)] = self.intervention_graph

        return copy.deepcopy(self, memo=memo)


class NNsightSamplingMetadata(SamplingMetadata):
    pass


from __future__ import annotations

import inspect
import io
import os
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import requests
import socketio
import torch
from tqdm.auto import tqdm

from ... import __IPYTHON__, CONFIG, __version__
from ..._c.py_mount import mount, unmount
from ...intervention.serialization import load, save
from ...log import remote_logger
from ...schema.request import RequestModel
from ...schema.response import RESULT, ResponseModel
from ..tracing.tracer import Tracer
from .base import Backend


class RemoteException(Exception):
    pass

class RemoteBackend(Backend):
    """Backend to execute a context object via a remote service.

    Context object must inherit from RemoteMixin and implement its methods.

    Attributes:

        url (str): Remote host url. Defaults to that set in CONFIG.API.HOST.
    """

    def __init__(
        self,
        model_key: str,
        host: str = None,
        blocking: bool = True,
        job_id: str = None,
        ssl: bool = None,
        api_key: str = "",
        callback: str = "",
    ) -> None:

        self.model_key = model_key
        
        self.host = host or os.environ.get("NDIF_HOST", None) or CONFIG.API.HOST
        self.api_key = api_key or os.environ.get("NDIF_API_KEY", None) or CONFIG.API.APIKEY

        self.job_id = job_id
        self.ssl = CONFIG.API.SSL if ssl is None else ssl
        self.zlib = CONFIG.API.ZLIB
        self.blocking = blocking
        self.callback = callback
        
        self.address = f"http{'s' if self.ssl else ''}://{self.host}"
        self.ws_address = f"ws{'s' if CONFIG.API.SSL else ''}://{self.host}"
        
        self.job_status = None

    def request(self, tracer: Tracer) -> Tuple[bytes, Dict[str, str]]:

        interventions = super().__call__(tracer)

        data = RequestModel(interventions=interventions, tracer=tracer).serialize(
            self.zlib
        )

        headers = {
            "nnsight-model-key": self.model_key,
            "nnsight-zlib": str(self.zlib),
            "nnsight-version": __version__,
            "ndif-api-key": self.api_key,
            "ndif-timestamp": str(time.time()),
            "ndif-callback": self.callback,
        }

        return data, headers

    def __call__(self, tracer = None):

        if self.blocking:

            # Do blocking request.
            result = self.blocking_request(tracer)

        else:

            # Otherwise we are getting the status / result of the existing job.
            result = self.non_blocking_request(tracer)

        if tracer is not None and result is not None:
            tracer.push(result)
            
        return result

    def handle_response(
        self, response: ResponseModel, tracer: Optional[Tracer] = None
    ) -> Optional[RESULT]:
        """Handles incoming response data.

        Logs the response object.
        If the job is completed, retrieve and stream the result from the remote endpoint.
        Use torch.load to decode and load the `ResultModel` into memory.
        Use the backend object's .handle_result method to handle the decoded result.

        Args:
            response (Any): Json data to concert to `ResponseModel`

        Raises:
            Exception: If the job's status is `ResponseModel.JobStatus.ERROR`

        Returns:
            ResponseModel: ResponseModel.
        """
        
        self.job_status = response.status
        
        if response.status == ResponseModel.JobStatus.ERROR:
            raise RemoteException(f"{response.description}\nRemote exception.")

        # Log response for user
        response.log(remote_logger)
        self.job_status = response.status

        # If job is completed:
        if response.status == ResponseModel.JobStatus.COMPLETED:

            # If the response has no result data, it was too big and we need to stream it from the server.
            if response.data is None:

                result = self.get_result(response.id)
            else:

                result = response.data
                
            return result
                
        elif response.status == ResponseModel.JobStatus.STREAM:
            
            model = getattr(tracer, "model", None)
            
            fn = load(response.data, model)
            
            local_tracer = LocalTracer(_info=tracer.info)
            
            local_tracer.execute(fn)

            


    def submit_request(
        self, data: bytes, headers: Dict[str, Any]
    ) -> Optional[ResponseModel]:
        """Sends request to the remote endpoint and handles the response object.

        Raises:
            Exception: If there was a status code other than 200 for the response.

        Returns:
            (ResponseModel): Response.
        """

        from ...schema.response import ResponseModel

        headers["Content-Type"] = "application/octet-stream"

        response = requests.post(
            f"{self.address}/request",
            data=data,
            headers=headers,
        )

        if response.status_code == 200:

            response = ResponseModel(**response.json())
            
            self.job_id = response.id

            self.handle_response(response)

            return response

        else:
            try:
                msg = response.json()["detail"]
            except:
                msg = response.reason
            raise ConnectionError(msg)

    def get_response(self) -> Optional[RESULT]:
        """Retrieves and handles the response object from the remote endpoint.

        Raises:
            Exception: If there was a status code other than 200 for the response.

        Returns:
            (ResponseModel): Response.
        """

        from ...schema.response import ResponseModel

        response = requests.get(
            f"{self.address}/response/{self.job_id}",
            headers={"ndif-api-key": self.api_key},
        )

        if response.status_code == 200:

            response = ResponseModel(**response.json())

            return self.handle_response(response)

        else:

            raise Exception(response.reason)

    def get_result(self, id: str) -> RESULT:

        result_bytes = io.BytesIO()
        result_bytes.seek(0)

        # Get result from result url using job id.
        with requests.get(
            url=f"{self.address}/result/{id}",
            stream=True,
        ) as stream:
            # Total size of incoming data.
            total_size = float(stream.headers["Content-length"])

            with tqdm(
                total=total_size,
                unit="B",
                unit_scale=True,
                desc="Downloading result",
            ) as progress_bar:
                # chunk_size=None so server determines chunk size.
                for data in stream.iter_content(chunk_size=None):
                    progress_bar.update(len(data))
                    result_bytes.write(data)

        # Move cursor to beginning of bytes.
        result_bytes.seek(0)

        # Decode bytes with pickle and then into pydantic object.
        result = torch.load(result_bytes, map_location="cpu", weights_only=False)

        # Close bytes
        result_bytes.close()

        return result

    def blocking_request(self, tracer: Tracer) -> Optional[RESULT]:
        """Send intervention request to the remote service while waiting for updates via websocket.

        Args:
            request (RequestModel):Request.
        """

        # We need to do some processing / optimizations on both the graph were sending remotely
        # and our local intervention graph. In order handle the more complex Protocols for streaming.

        # Create a socketio connection to the server.
        with socketio.SimpleClient(reconnection_attempts=10) as sio:
            # Connect
            sio.connect(
                self.ws_address,
                socketio_path="/ws/socket.io",
                transports=["websocket"],
                wait_timeout=10,
            )

            data, headers = self.request(tracer)

            headers["ndif-session_id"] = sio.sid

            # Submit request via
            response = self.submit_request(data, headers)

            try:
                LocalTracer.register(lambda data: self.stream_send(data, sio))
                # Loop until
                while True:

                    # Get pickled bytes value from the websocket.
                    response = sio.receive()[1]
                    # Convert to pydantic object.
                    response = ResponseModel.unpickle(response)
                    # Handle the response.
                    result = self.handle_response(response, tracer=tracer)
                    # Break when completed.
                    if result is not None:
                        return result

            except Exception as e:

                raise e
            
            finally:
                LocalTracer.deregister()
        

    def stream_send(
        self, values: Dict[int, Any],  sio: socketio.SimpleClient
    ):
        """Upload some value to the remote service for some job id.

        Args:
            value (Any): Value to upload
            job_id (str): Job id.
            sio (socketio.SimpleClient): Connected websocket client.
        """
        
        data = save(values)

        sio.emit(
            "stream_upload",
            data=(data, self.job_id),
        )

    def non_blocking_request(self, tracer: Tracer):
        """Send intervention request to the remote service if request provided. Otherwise get job status.

        Sets CONFIG.API.JOB_ID on initial request as to later get the status of said job.

        When job is completed, clear CONFIG.API.JOB_ID to request a new job.

        Args:
            request (RequestModel): Request if submitting a new request. Defaults to None
        """

        if self.job_id is None:

            data, headers = self.request(tracer)

            # Submit request via
            response = self.submit_request(data, headers)

            self.job_id = response.id
            
        else:

            result = self.get_response()

            return result



class LocalTracer(Tracer):
    
    _send: Callable = None
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.remotes = set()
        
    @classmethod
    def register(cls, send_fn: Callable):
        
        cls._send = send_fn
        
    @classmethod
    def deregister(cls):
        
        cls._send = None
    
    def _save_remote(self, obj: Any):
        
        self.remotes.add(id(obj))
        
    def execute(self, fn: Callable):
        
        mount(self._save_remote, "remote")
        
        fn(self, self.info)
        
        unmount("remote")
        
        return 
        
    def push(self):
        
        # Find the frame where the traced code is executing
        state_frame = inspect.currentframe().f_back

        state = state_frame.f_locals
        
        super().push(state)
        
        state = {k:v for k,v in state.items() if id(v) in self.remotes}

        LocalTracer._send(state)



import pytest
import torch
import PIL

if not torch.cuda.is_available():
    pytest.skip("no GPU available.", allow_module_level=True)

from nnsight.modeling.diffusion import DiffusionModel

@pytest.fixture(scope="module")
def tiny_sd():
    return DiffusionModel("segmind/tiny-sd", torch_dtype=torch.float16, dispatch=True).to("cuda")

@pytest.fixture(scope="module")
def cat_prompt():
    return "A brown and white cat staring off with pretty green eyes"

def test_generation(tiny_sd, cat_prompt):
    num_images_per_prompt = 3

    images = tiny_sd.generate(cat_prompt, 
                              num_inference_steps=20, 
                              num_images_per_prompt=num_images_per_prompt,
                              seed=423, 
                              trace=False).images

    assert len(images) == num_images_per_prompt
    assert all([type(img) == PIL.Image.Image for img in images])
    assert all(images[i] != images[j] for i in range(num_images_per_prompt) for j in range(i + 1, num_images_per_prompt))    



from typing import Callable, TYPE_CHECKING, Any, Union
from .base import Tracer
from ..interleaver import Interleaver, Mediator
from .util import try_catch

class IteratorProxy:
    
    def __init__(self, interleaver: Interleaver):
        self.interleaver = interleaver
        
    def __getitem__(self, iteration: Union[int, slice]):
        return IteratorTracer(iteration, self.interleaver)
    
class IteratorTracer(Tracer):
    
    def __init__(self, iteration: Union[int, slice], interleaver: Interleaver):
        super().__init__()
        
        self.interleaver = interleaver
        
        self.iteration = iteration
        
    def compile(self):
        """
        Compile the captured source code as a callable function.

        Wraps the captured code in a function definition that accepts the
        necessary context parameters for execution.

        Returns:
            A callable function that executes the captured code block
        """

        iteration_var_name = self.info.node.items[0].optional_vars.id if self.info.node.items[0].optional_vars is not None else "__nnsight_iteration__"

        # Wrap the captured code in a function definition with appropriate parameters
        self.info.source = [
            f"def __nnsight_tracer_{id(self)}__(__nnsight_mediator__, __nnsight_tracing_info__, {iteration_var_name}):\n",
            "    __nnsight_mediator__.pull()\n",
            *try_catch(
                self.info.source,
                exception_source=["__nnsight_mediator__.exception(exception)\n"],
                else_source=["__nnsight_mediator__.end()\n"],
            ),
        ]
        
        self.info.start_line -= 2
        
    def execute(self, fn: Callable):
        
        mediator = Mediator(fn, self.info, batch_group=self.interleaver.current.batch_group, stop=self.interleaver.current.all_stop)

        mediator.name = "Iterator" + mediator.name
        
        self.interleaver.current.iter(mediator, self.iteration)
    
    
    


from vllm.v1.worker import gpu_worker
from ..model_runners.GPUModelRunner import NNsightGPUModelRunner


class NNsightGPUWorker(gpu_worker.Worker):

    def __init__(self, *args, **kwargs):
        
        gpu_worker.GPUModelRunner = NNsightGPUModelRunner
        
        super().__init__(*args, **kwargs)


from builtins import compile, exec
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..tracing.tracer import Tracer
else:
    Tracer = Any
    

class Backend:
    
    def __call__(self, tracer: Tracer):
        
        
        tracer.compile()

        source = "".join(tracer.info.source)
        
        code_obj = compile(source, tracer.info.filename, "exec")
        
        local_namespace = {}

        # Execute the function definition in the local namespace
        exec(
            code_obj,
            {**tracer.info.frame.f_globals, **tracer.info.frame.f_locals},
            local_namespace,
        )

        fn = list(local_namespace.values())[-1]
        
        return fn


# This section ensures that the source code for both the main script and the file where `nnsight` is imported
# is cached in Python's `linecache` module. This is important for robust stack trace and debugging support,
# especially in interactive or dynamic environments where files may change after import.
# 
# 1. We first attempt to cache the main script (the file passed to the Python interpreter) by calling
#    `linecache.getlines` on `sys.argv[0]`. This ensures that if the script is modified or deleted after
#    execution starts, traceback and inspection tools can still access its source.
# 2. Next, we walk up the call stack to find the first frame outside of importlib (i.e., the user code that
#    imported `nnsight`), and cache its source file as well. This helps ensure that the file where `nnsight`
#    is imported is also available in `linecache`, even if it changes later.

import linecache
import sys
import os

try:
    # Cache the main script file
    linecache.getlines(os.path.abspath(sys.argv[0]))
except Exception:
    pass

import inspect

try:
    # Walk up the stack to cache the file where nnsight is imported
    frame = inspect.currentframe()
    while frame.f_back:
        frame = frame.f_back
        if 'importlib' not in frame.f_code.co_filename:
            linecache.getlines(frame.f_code.co_filename, frame.f_globals)
except Exception:
    pass

from functools import wraps

import os, yaml
import warnings
from typing import Optional
from .schema.config import ConfigModel

PATH = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(PATH, "config.yaml"), "r") as file:
    CONFIG = ConfigModel(**yaml.safe_load(file))
    
from importlib.metadata import PackageNotFoundError, version
  
try:
    __version__ = version("nnsight")
except PackageNotFoundError:
    __version__ = "unknown version"

from IPython import get_ipython

try:
    __IPYTHON__ = get_ipython() is not None
except NameError:
    __IPYTHON__ = False
    
__INTERACTIVE__ = (sys.flags.interactive or not sys.argv[0]) and not __IPYTHON__

base_deprecation_message = "is deprecated as of v0.5.0 and will be removed in a future version."

NNS_VLLM_VERSION = "0.9.2"
    
def deprecated(message:Optional[str]=None, error:bool=False):
    
    def decorator(func):
        
        @wraps(func)
        def inner(*args, **kwargs):
            
            deprecation_message = (
                f"{func.__module__}.{func.__name__} {base_deprecation_message}"
                + (f"\n{message}" if message is not None else "")
            )
            
            if error:
                raise DeprecationWarning(deprecation_message)
            else:
                warnings.warn(deprecation_message)
                
            return func(*args, **kwargs)
        
        return inner
    
    return decorator




from .intervention.envoy import Envoy
from .modeling.base import NNsight
from .modeling.language import LanguageModel
from .intervention.tracing.base import Tracer


    
def session(*args, **kwargs):
    return Tracer(*args, **kwargs)

from .util import Patcher, Patch

DEFAULT_PATCHER = Patcher()

# Tensor creation operations
from torch._subclasses.fake_tensor import FakeTensor


def fake_bool(self):
    return True


DEFAULT_PATCHER.add(Patch(FakeTensor, fake_bool, "__bool__"))

from torch.amp.autocast_mode import autocast

def wrap_autocast(func):
    
    @wraps(func)
    def inner(self, device_type:str, *args, **kwargs):
        
        if device_type == "meta":
            device_type = "cpu"
            
        return func(self, device_type, *args, **kwargs)
        
    return inner


DEFAULT_PATCHER.add(Patch(autocast, wrap_autocast(autocast.__init__), "__init__"))


from torch import Tensor
from .intervention.tracing.backwards import BackwardsTracer
from .intervention.tracing.base import WithBlockNotFoundError

def wrap_backward(func):

    @wraps(func)
    def inner(tensor: Tensor, *args, **kwargs):
        
        try:

            tracer = BackwardsTracer(tensor, func, *args, **kwargs)

        except WithBlockNotFoundError:

            return func(tensor, *args, **kwargs)

        return tracer

    return inner

DEFAULT_PATCHER.add(Patch(Tensor, wrap_backward(Tensor.backward), "backward"))

DEFAULT_PATCHER.__enter__()


## TODO: Legacy

@deprecated()
def apply(fn, *args, **kwargs):
    
    return fn(*args, **kwargs)

@deprecated()
def log(message:str):
    
    print(message)
  
@deprecated(error=True)
def local(*args, **kwargs):
    pass

@deprecated(error=True)
def cond(*args, **kwargs):
    pass

@deprecated(error=True)
def iter(*args, **kwargs):
    pass

from .intervention.interleaver import EarlyStopException

@deprecated()
def stop():
    raise EarlyStopException()

@deprecated()
def trace(fn):
    
    return fn


bool = deprecated(message="Use the standard `bool()` instead.")(bool)
bytes = deprecated(message="Use the standard `bytes()` instead.")(bytes)
complex = deprecated(message="Use the standard `complex()` instead.")(complex)
dict = deprecated(message="Use the standard `dict()` instead.")(dict)
float = deprecated(message="Use the standard `float()` instead.")(float)
int = deprecated(message="Use the standard `int()` instead.")(int)
list = deprecated(message="Use the standard `list()` instead.")(list)
str = deprecated(message="Use the standard `str()` instead.")(str)
tuple = deprecated(message="Use the standard `tuple()` instead.")(tuple)
bytearray = deprecated(message="Use the standard `bytearray()` instead.")(bytearray)

if __INTERACTIVE__:
    from code import InteractiveConsole
    import readline
    
    # We need to use our own console so when we trace, we can get the source code from the buffer of our custom console.
    class NNsightConsole(InteractiveConsole):
        pass
    
    # We want the new console to have the locals of the interactive frame the user is already in.
    frame = inspect.currentframe()

    while frame.f_back:
        frame = frame.f_back

    # This was kicked off by importing nnsight, but upon entering the new console, we wont actually have the import in the new locals.
    # So we need to get the last command from the history and execute it in the new locals.
    length = readline.get_current_history_length()
    l1 = readline.get_history_item(length)
    ilocals = {**frame.f_locals}
    exec(l1, frame.f_globals, ilocals)
    
    __INTERACTIVE_CONSOLE__ = NNsightConsole(filename="<nnsight-console>", locals=ilocals)
    __INTERACTIVE_CONSOLE__.interact(banner="")

else:
    __INTERACTIVE_CONSOLE__ = None



import io
import pickle
from builtins import open
from types import FrameType
from typing import TYPE_CHECKING, Any, Optional, Union

import cloudpickle

from .envoy import Envoy


class CustomCloudPickler(cloudpickle.Pickler):
    def persistent_id(self, obj):

        if isinstance(obj, Envoy):
            return f"ENVOY:{obj.path}"

        if isinstance(obj, FrameType):
            return "FRAME"

        return None


class CustomCloudUnpickler(pickle.Unpickler):
    def __init__(self, file, root: Envoy, frame: FrameType):
        super().__init__(file)
        self.root = root
        self.frame = frame

    def persistent_load(self, pid):

        if pid.startswith("ENVOY:"):
            path = pid.removeprefix("ENVOY:model")
            return self.root.get(path)

        if pid == "FRAME":
            return self.frame

        raise pickle.UnpicklingError(f"Unknown persistent id: {pid}")


def save(obj: Any, path: Optional[str] = None):


    if path is None:
        file = io.BytesIO()
        CustomCloudPickler(file, protocol=4).dump(obj)
        file.seek(0)
        return file.read()

    with open(path, "wb") as file:
        CustomCloudPickler(file).dump(obj)


def load(data: Union[str, bytes], model: Envoy, frame: Optional[FrameType] = None):

    if isinstance(data, bytes):
        return CustomCloudUnpickler(io.BytesIO(data), model, frame).load()

    with open(data, "rb") as file:
        return CustomCloudUnpickler(file, model, frame).load()



from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, Set

import torch
import torch.distributed

import vllm.envs as envs
from vllm.distributed.kv_transfer import (get_kv_transfer_group,
                                          has_kv_transfer_group)
from vllm.distributed.parallel_state import get_pp_group, get_tp_group
from vllm.forward_context import set_forward_context
from vllm.model_executor.layers.rotary_embedding import MRotaryEmbedding
from vllm.sampling_params import SamplingType
from vllm.sequence import IntermediateTensors
from vllm.utils import LazyLoader, round_up
from vllm.v1.outputs import EMPTY_MODEL_RUNNER_OUTPUT, ModelRunnerOutput
from vllm.v1.worker.gpu_input_batch import CachedRequestState
from vllm.v1.worker.gpu_model_runner import GPUModelRunner

if TYPE_CHECKING:

    from vllm.v1.core.sched.output import SchedulerOutput, NewRequestData

    from ....intervention.interleaver import Interleaver
else:
    xgr = LazyLoader("xgr", globals(), "xgrammar")

class NNsightGPUModelRunner(GPUModelRunner):

    class NNsightRequestHelper:
        '''
        Helper class for batching requests in the GPUModelRunner.

        Attributes:
            ids_to_batch_group (Dict[str, int]): Dictionary mapping request IDs to their assigned batch group indices.
            interleaver_to_ids (Dict[Interleaver, Set[str]]): Dictionary mapping interleavers to sets of request IDs.
            flat_batch_groups (Dict[Interleaver, List[Tuple[int, int]]]): Dictionary mapping interleavers to their flattened batch groups.

        Methods:
            process_new_reqs(new_reqs: List[NewRequestData]) -> None: Process new requests and compute the flat batch groups.
            process_finished_req(req_id: str, interleaver: Interleaver) -> None: Process a finished request,
                by updating batch groups and cleaning up mappings.
        '''

        def __init__(self):

            self.ids_to_batch_group: Dict[str, int] = {}
            self.interleaver_to_ids: Dict["Interleaver", Set[str]] = defaultdict(set)
            self.flat_batch_groups: Dict["Interleaver", List[Tuple[int, int]]] = defaultdict(list)

        def process_new_reqs(self, new_reqs: List["NewRequestData"]) -> None:
            """
            Process new requests and organize them into batch groups for execution.
            
            This method handles the batching logic for new requests, organizing them
            into appropriate batch groups based on their interleaver's batching strategy.
            
            Args:
                new_reqs (List[NewRequestData]): List of new request data objects to process.
                    Each request contains sampling parameters with an associated interleaver
                    that defines the batching behavior.
            
            Notes:
                - Resets the flat_batch_groups dictionary at the start
                - For interleavers that require batching, requests are assigned to batch groups
                - Batch groups are tuples of (start_position, size) indicating token ranges
                - Updates internal tracking dictionaries for request-to-batch-group mapping
                - Advances to next batch group when current group capacity is exceeded
            """

            # reset
            self.flat_batch_groups = defaultdict(list)

            # current batch group index per interleaver
            curr_batch_groups: Dict["Interleaver", int] = defaultdict(int)

            interleavers: Set["Interleaver"] = set()

            for new_req in new_reqs:
                interleaver = new_req.sampling_params.interleaver

                # clean up any finished requests from the previous round
                if interleaver not in interleavers:
                    interleavers.add(interleaver)

                    if interleaver in self.interleaver_to_ids:
                        for req_id in self.interleaver_to_ids[interleaver]:
                            self.ids_to_batch_group.pop(req_id)

                    self.interleaver_to_ids[interleaver] = set()

                if interleaver.batcher.needs_batching:
                    # linking request to batch group
                    batch_groups: List[Tuple[int, int]] = interleaver.batcher.batch_groups
                    batch_group_idx: int = curr_batch_groups[interleaver]

                    self.interleaver_to_ids[interleaver].add(new_req.req_id)

                    # update current batch group index
                    if len(self.interleaver_to_ids[interleaver]) > sum(batch_groups[batch_group_idx]):
                        curr_batch_groups[interleaver] += 1

                    self.ids_to_batch_group[new_req.req_id] = curr_batch_groups[interleaver]

                    # first seen request
                    if self.flat_batch_groups[interleaver] == []:
                        # req_to_interleaver_dict[interleaver] = (0, 0)
                        # the first batch group starts at 0 and has a size of the number of tokens in the prompt
                        batch_group = (batch_groups[0][0], len(new_req.prompt_token_ids))
                        self.flat_batch_groups[interleaver].append(batch_group)

                    else:
                        req_count: int = len(self.interleaver_to_ids[interleaver])
                        curr_batch_group: Tuple[int, int] = batch_groups[batch_group_idx]
                        last_flat_batch_group: Tuple[int, int] = self.flat_batch_groups[interleaver][-1]

                        # check if the current request is within the current batch group
                        if req_count < sum(curr_batch_group):
                            # update the current batch group with the number of tokens in the new request
                            batch_group = (last_flat_batch_group[0], last_flat_batch_group[1] + len(new_req.prompt_token_ids))
                            self.flat_batch_groups[interleaver][-1] = batch_group

                        else:
                            # add a new batch group
                            batch_group = (sum(last_flat_batch_group), len(new_req.prompt_token_ids))
                            self.flat_batch_groups[interleaver].append(batch_group)

        def process_finished_req(self, req_id: str, interleaver: "Interleaver") -> None:
            """
            Process a finished request by updating batch groups and cleaning up mappings.
            
            This method handles the removal of a completed request from the batch tracking
            system. When a request finishes, it:
            1. Removes the request from the interleaver's tracking set
            2. Decrements the batch group count 
            3. If the batch group becomes empty, removes it and updates indices
            4. Updates batch group mappings for remaining requests
            5. Cleans up the request ID mapping
            
            Args:
                req_id (str): The unique identifier of the finished request
                interleaver (Interleaver): The interleaver instance managing this request
            """

            if req_id in self.interleaver_to_ids[interleaver]:

                # remove finished request
                self.interleaver_to_ids[interleaver].remove(req_id)
                
                batch_idx = self.ids_to_batch_group[req_id]
                batch_groups = interleaver.batcher.batch_groups

                batch_group = batch_groups[batch_idx]
                new_batch_group = (batch_group[0], batch_group[1] - 1)

                if new_batch_group[1] == 0:
                    batch_groups.pop(batch_idx)

                    for idx in range(batch_idx, len(batch_groups)):
                        # update the batch start
                        if idx == batch_idx:
                            batch_groups[idx] = (batch_group[0], batch_groups[idx][1])
                        else:
                            batch_groups[idx] = (sum(batch_groups[idx-1]), batch_groups[idx][1])

                    # update the batch group index for remaining requests
                    for other_req_id in self.interleaver_to_ids[interleaver]:
                        if self.ids_to_batch_group[other_req_id] > batch_idx:
                            self.ids_to_batch_group[other_req_id] -= 1

                    for mediator in interleaver.invokers:
                        if mediator.batch_group and mediator.batch_group > batch_idx:
                            mediator.batch_group -= 1

                        if mediator.child:
                            mediator.child.batch_group = mediator.batch_group

                    interleaver.batcher.batch_groups = batch_groups
                        
                else:
                    batch_groups[batch_idx] = new_batch_group

                # remove finished request
                self.ids_to_batch_group.pop(req_id)


    def __init__(self, *args, **kwargs):
        
        from .. import VLLM
        
        super().__init__(*args, **kwargs)
        
        self.nnsight_model: VLLM

        self.nnsight_request_helper = self.NNsightRequestHelper()

    def load_model(self) -> None:
        
        from .. import VLLM

        super().load_model()

        self.nnsight_model = VLLM(self.model)

    def _update_states(self, scheduler_output: "SchedulerOutput") -> None:
        """Update the cached states and the persistent batch with the scheduler
        output.

        The updated states are used by the `_prepare_inputs` function to create
        the input GPU tensors for the model.

        The SamplingMetadata is updated and copied to the GPU if there is a
        new/resumed/paused/finished request in the batch.
        """

        ########## NNsight ##########

        self.nnsight_request_helper.process_new_reqs(scheduler_output.scheduled_new_reqs)

        #############################

        # Remove finished requests from the cached states.
        for req_id in scheduler_output.finished_req_ids:
            
            ########## NNsight ##########

            self.nnsight_request_helper.process_finished_req(
                req_id, 
                self.requests[req_id].sampling_params.interleaver
            )

            #############################

            self.requests.pop(req_id, None)
            self.encoder_cache.pop(req_id, None)
        # Remove the finished requests from the persistent batch.
        # NOTE(woosuk): There could be an edge case where finished_req_ids and
        # scheduled_req_ids overlap. This happens when a request is aborted and
        # then resubmitted with the same ID. In this case, we treat them as two
        # distinct requests - clearing the cached states for the first request
        # and handling the second as a new request.
        for req_id in scheduler_output.finished_req_ids:
            self.input_batch.remove_request(req_id)

        # Free the cached encoder outputs.
        for req_id, input_id in scheduler_output.free_encoder_input_ids:
            encoder_outputs = self.encoder_cache.get(req_id)
            if encoder_outputs is not None:
                encoder_outputs.pop(input_id, None)
                if not encoder_outputs:
                    self.encoder_cache.pop(req_id, None)

        # Remove the unscheduled requests from the persistent batch.
        # NOTE(woosuk): The unscheduled requests are either preempted requests
        # or running requests that are not scheduled in this step. We remove
        # them from the persistent batch but keep their cached states since
        # they will be scheduled again sometime in the future.
        scheduled_req_ids = scheduler_output.num_scheduled_tokens.keys()
        cached_req_ids = self.input_batch.req_id_to_index.keys()
        unscheduled_req_ids = cached_req_ids - scheduled_req_ids
        # NOTE(woosuk): The persistent batch optimization assumes that
        # consecutive batches contain mostly the same requests. If batches
        # have low request overlap (e.g., alternating between two distinct
        # sets of requests), this optimization becomes very inefficient.
        for req_id in unscheduled_req_ids:
            self.input_batch.remove_request(req_id)

        req_ids_to_add: list[str] = []

        # Add new requests to the cached states.
        for new_req_data in scheduler_output.scheduled_new_reqs:
            req_id = new_req_data.req_id
            sampling_params = new_req_data.sampling_params

            ########## NNsight ##########

            # flatten batch groups to match vLLM's batch processing method
            batcher = sampling_params.interleaver.batcher
            batcher.cache_batch_groups(self.nnsight_request_helper.flat_batch_groups[sampling_params.interleaver])

            #############################

            pooling_params = new_req_data.pooling_params
            if sampling_params and \
                sampling_params.sampling_type == SamplingType.RANDOM_SEED:
                generator = torch.Generator(device=self.device)
                generator.manual_seed(sampling_params.seed)
            else:
                generator = None

            self.requests[req_id] = CachedRequestState(
                req_id=req_id,
                prompt_token_ids=new_req_data.prompt_token_ids,
                mm_inputs=new_req_data.mm_inputs,
                mm_positions=new_req_data.mm_positions,
                sampling_params=sampling_params,
                pooling_params=pooling_params,
                generator=generator,
                block_ids=new_req_data.block_ids,
                num_computed_tokens=new_req_data.num_computed_tokens,
                output_token_ids=[],
                lora_request=new_req_data.lora_request,
            )

            # Only relevant for models using M-RoPE (e.g, Qwen2-VL)
            if self.uses_mrope:
                image_grid_thw = []
                video_grid_thw = []
                second_per_grid_ts = []
                audio_feature_lengths = []
                use_audio_in_video = False
                for mm_input in self.requests[req_id].mm_inputs:
                    if mm_input.get("image_grid_thw") is not None:
                        image_grid_thw.extend(
                            mm_input["image_grid_thw"].tolist())
                    if mm_input.get("video_grid_thw") is not None:
                        video_grid_thw.extend(
                            mm_input["video_grid_thw"].tolist())
                    if mm_input.get("second_per_grid_ts") is not None:
                        second_per_grid_ts.extend(
                            mm_input["second_per_grid_ts"])
                    if mm_input.get("audio_feature_lengths") is not None:
                        audio_feature_lengths.extend(
                            mm_input["audio_feature_lengths"])
                    if mm_input.get("use_audio_in_video") is True:
                        use_audio_in_video = True

                hf_config = self.model_config.hf_config

                self.requests[req_id].mrope_positions, \
                    self.requests[req_id].mrope_position_delta = \
                    MRotaryEmbedding.get_input_positions_tensor(
                        self.requests[req_id].prompt_token_ids,
                        hf_config=hf_config,
                        image_grid_thw=image_grid_thw,
                        video_grid_thw=video_grid_thw,
                        second_per_grid_ts=second_per_grid_ts,
                        audio_feature_lengths=audio_feature_lengths,
                        use_audio_in_video=use_audio_in_video,
                    )

            req_ids_to_add.append(req_id)

        # Update the states of the running/resumed requests.
        is_last_rank = get_pp_group().is_last_rank
        req_data = scheduler_output.scheduled_cached_reqs
        for i, req_id in enumerate(req_data.req_ids):
            req_state = self.requests[req_id]
            num_computed_tokens = req_data.num_computed_tokens[i]
            new_block_ids = req_data.new_block_ids[i]
            resumed_from_preemption = req_data.resumed_from_preemption[i]

            # Update the cached states.
            req_state.num_computed_tokens = num_computed_tokens

            if not is_last_rank:
                # When using PP, the scheduler sends the sampled tokens back,
                # because there's no direct communication between the first-
                # stage worker and the last-stage worker.
                new_token_ids = req_data.new_token_ids[i]
                # Add the sampled token(s) from the previous step (if any).
                # This doesn't include "unverified" tokens like spec tokens.
                num_new_tokens = (num_computed_tokens + len(new_token_ids) -
                                  req_state.num_tokens)
                if num_new_tokens == 1:
                    # Avoid slicing list in most common case.
                    req_state.output_token_ids.append(new_token_ids[-1])
                elif num_new_tokens > 0:
                    req_state.output_token_ids.extend(
                        new_token_ids[-num_new_tokens:])

            # Update the block IDs.
            if not resumed_from_preemption:
                # Append the new blocks to the existing block IDs.
                for block_ids, new_ids in zip(req_state.block_ids,
                                              new_block_ids):
                    block_ids.extend(new_ids)
            else:
                # The request is resumed from preemption.
                # Replace the existing block IDs with the new ones.
                req_state.block_ids = new_block_ids

            req_index = self.input_batch.req_id_to_index.get(req_id)
            if req_index is None:
                # The request is not in the persistent batch.
                # The request was either preempted and resumed later, or was not
                # scheduled in the previous step and needs to be added again.
                req_ids_to_add.append(req_id)
                continue

            # Update the persistent batch.
            self.input_batch.num_computed_tokens_cpu[req_index] = (
                num_computed_tokens)
            self.input_batch.block_table.append_row(new_block_ids, req_index)

            # For the last rank, we don't need to update the token_ids_cpu
            # because the sampled tokens are already cached.
            if not is_last_rank:
                # Add new_token_ids to token_ids_cpu.
                start_token_index = num_computed_tokens
                end_token_index = num_computed_tokens + len(new_token_ids)
                self.input_batch.token_ids_cpu[
                    req_index,
                    start_token_index:end_token_index] = new_token_ids
                self.input_batch.num_tokens_no_spec[
                    req_index] = end_token_index
                self.input_batch.num_tokens[req_index] = end_token_index

            # Add spec_token_ids to token_ids_cpu.
            spec_token_ids = (
                scheduler_output.scheduled_spec_decode_tokens.get(req_id, ()))
            if spec_token_ids:
                num_spec_tokens = len(spec_token_ids)
                start_index = self.input_batch.num_tokens_no_spec[req_index]
                end_token_index = start_index + num_spec_tokens
                self.input_batch.token_ids_cpu[
                    req_index, start_index:end_token_index] = spec_token_ids
                # NOTE(woosuk): `num_tokens` here may include spec tokens.
                self.input_batch.num_tokens[req_index] += num_spec_tokens

        # Add the new or resumed requests to the persistent batch.
        # The smaller empty indices are filled first.
        for req_id in req_ids_to_add:
            req_state = self.requests[req_id]
            self.input_batch.add_request(req_state)

        # Condense the batched states if there are gaps left by removed requests
        self.input_batch.condense()
        # Allow attention backend to reorder the batch, potentially
        self._may_reorder_batch(scheduler_output)
        # Refresh batch metadata with any pending updates.
        self.input_batch.refresh_metadata()


    @torch.inference_mode()
    def execute_model(
        self,
        scheduler_output: "SchedulerOutput",
        intermediate_tensors: Optional[IntermediateTensors] = None,
    ) -> Union[ModelRunnerOutput, IntermediateTensors]:

        self._update_states(scheduler_output)
        if not scheduler_output.total_num_scheduled_tokens:
            if not has_kv_transfer_group():
                # Return empty ModelRunnerOutput if there's no work to do.
                return EMPTY_MODEL_RUNNER_OUTPUT

            return self.kv_connector_no_forward(scheduler_output)

        # Prepare the decoder inputs.
        (attn_metadata, attention_cuda_graphs, logits_indices,
         spec_decode_metadata,
         num_scheduled_tokens_np) = (self._prepare_inputs(scheduler_output))
        num_scheduled_tokens = scheduler_output.total_num_scheduled_tokens
        if (self.use_cuda_graph
                and num_scheduled_tokens <= self.cudagraph_batch_sizes[-1]):
            # Use piecewise CUDA graphs.
            # Add padding to the batch size.
            num_input_tokens = self.vllm_config.pad_for_cudagraph(
                num_scheduled_tokens)
        else:
            # Eager mode.
            # Pad tokens to multiple of tensor_parallel_size when
            # enabled collective fusion for SP
            tp_size = self.vllm_config.parallel_config.tensor_parallel_size
            if self.compilation_config.pass_config. \
                enable_sequence_parallelism and tp_size > 1:
                num_input_tokens = round_up(num_scheduled_tokens, tp_size)
            else:
                num_input_tokens = num_scheduled_tokens

        # Padding for DP
        num_pad, num_tokens_across_dp = self.get_dp_padding(num_input_tokens)
        num_input_tokens += num_pad

        # _prepare_inputs may reorder the batch, so we must gather multi
        # modal outputs after that to ensure the correct order
        if self.is_multimodal_model:
            # Run the multimodal encoder if any.
            self._execute_mm_encoder(scheduler_output)
            mm_embeds = self._gather_mm_embeddings(scheduler_output)
        else:
            mm_embeds = []

        if self.is_multimodal_model and get_pp_group().is_first_rank:
            # NOTE(woosuk): To unify token ids and soft tokens (vision
            # embeddings), we always use embeddings (rather than token ids)
            # as input to the multimodal model, even when the input is text.
            input_ids = self.input_ids[:num_scheduled_tokens]
            if mm_embeds:
                inputs_embeds = self.model.get_input_embeddings(
                    input_ids, mm_embeds)
            else:
                inputs_embeds = self.model.get_input_embeddings(input_ids)
            # TODO(woosuk): Avoid the copy. Optimize.
            self.inputs_embeds[:num_scheduled_tokens].copy_(inputs_embeds)
            inputs_embeds = self.inputs_embeds[:num_input_tokens]
            input_ids = None
        else:
            # For text-only models, we use token ids as input.
            # While it is possible to use embeddings as input just like the
            # multimodal models, it is not desirable for performance since
            # then the embedding layer is not included in the CUDA graph.
            input_ids = self.input_ids[:num_input_tokens]
            inputs_embeds = None
        if self.uses_mrope:
            positions = self.mrope_positions[:, :num_input_tokens]
        else:
            positions = self.positions[:num_input_tokens]

        if get_pp_group().is_first_rank:
            intermediate_tensors = None
        else:
            intermediate_tensors = self.sync_and_slice_intermediate_tensors(
                num_input_tokens, intermediate_tensors, True)

        # Some attention backends only support CUDA Graphs in pure decode.
        # If attention doesn't support CUDA Graphs for this batch, but we
        # compiled with full CUDA graphs, we have to skip them entirely.
        skip_cuda_graphs = self.full_cuda_graph and not attention_cuda_graphs

        ### NNsight ####
        from contextlib import ExitStack

        interleavers = set([request.sampling_params.interleaver for request in  self.requests.values()])
        
        with ExitStack() as stack:
            for interleaver in interleavers:
                stack.enter_context(interleaver)

            # Run the model.
            # Use persistent buffers for CUDA graphs.
            with set_forward_context(
                    attn_metadata,
                    self.vllm_config,
                    num_tokens=num_input_tokens,
                    num_tokens_across_dp=num_tokens_across_dp,
                    skip_cuda_graphs=skip_cuda_graphs,
            ):
                self.maybe_setup_kv_connector(scheduler_output)

                model_output = self.model(
                    input_ids=input_ids,
                    positions=positions,
                    intermediate_tensors=intermediate_tensors,
                    inputs_embeds=inputs_embeds,
                )

                self.maybe_wait_for_kv_save()
                finished_sending, finished_recving = (
                    self.get_finished_kv_transfers(scheduler_output))
            

            if self.use_aux_hidden_state_outputs:
                hidden_states, aux_hidden_states = model_output
            else:
                hidden_states = model_output
                aux_hidden_states = None

            # Broadcast PP output for external_launcher (torchrun)
            # to make sure we are synced across pp ranks
            # TODO: Support overlapping mirco-batches
            # https://github.com/vllm-project/vllm/issues/18019
            broadcast_pp_output = \
                self.parallel_config.distributed_executor_backend \
                == "external_launcher" and len(get_pp_group().ranks) > 0
            if not get_pp_group().is_last_rank:
                # For mid-pipeline stages, return the hidden states.
                if not broadcast_pp_output:
                    return hidden_states
                assert isinstance(hidden_states, IntermediateTensors)
                get_pp_group().send_tensor_dict(hidden_states.tensors,
                                                all_gather_group=get_tp_group())
                logits = None
            else:
                if self.input_batch.pooling_params:
                    return self._pool(hidden_states, num_scheduled_tokens,
                                    num_scheduled_tokens_np, finished_sending,
                                    finished_recving)

                sample_hidden_states = hidden_states[logits_indices]
                logits = self.model.compute_logits(sample_hidden_states, None)
            if broadcast_pp_output:
                model_output_broadcast_data = {
                    "logits": logits.contiguous(),
                } if logits is not None else {}
                model_output_broadcast_data = get_pp_group().broadcast_tensor_dict(
                    model_output_broadcast_data, src=len(get_pp_group().ranks) - 1)
                assert model_output_broadcast_data is not None
                logits = model_output_broadcast_data["logits"]

            # Apply structured output bitmasks if present
            if scheduler_output.grammar_bitmask is not None:
                self.apply_grammar_bitmask(scheduler_output, logits)


            #### NNsight ####

            # resetting the batch_groups on the interleaver's batcher
            for interleaver in interleavers:
                interleaver.batcher.restore_batch_groups()

            logits = self.model.logits(logits, hook=True)

            ###############


            # Sample the next token and get logprobs if needed.
            sampling_metadata = self.input_batch.sampling_metadata
            if spec_decode_metadata is None:
                sampler_output = self.sampler(
                    logits=logits,
                    sampling_metadata=sampling_metadata,
                )
            else:
                # When indexing with a tensor (bonus_logits_indices), PyTorch
                # creates a new tensor with separate storage from the original
                # logits tensor. This means any in-place operations on bonus_logits
                # won't affect the original logits tensor.
                assert logits is not None
                bonus_logits = logits[spec_decode_metadata.bonus_logits_indices]
                sampler_output = self.sampler(
                    logits=bonus_logits,
                    sampling_metadata=sampling_metadata,
                )
                bonus_token_ids = sampler_output.sampled_token_ids

                # Just like `bonus_logits`, `target_logits` is a new tensor with
                # separate storage from the original `logits` tensor. Therefore,
                # it is safe to update `target_logits` in place.
                target_logits = logits[spec_decode_metadata.target_logits_indices]
                output_token_ids = self.rejection_sampler(
                    spec_decode_metadata,
                    None,  # draft_probs
                    target_logits,
                    bonus_token_ids,
                    sampling_metadata,
                )
                sampler_output.sampled_token_ids = output_token_ids


            ##### NNsight #####

            samples = self.model.samples(sampler_output.sampled_token_ids, hook=True)

            ###################

        ###################


        ####### NNsight #######

        for interleaver in interleavers:
                interleaver.invokers = [invoker for invoker in interleaver.invokers if invoker.alive]

        #######################


        num_nans_in_logits = {}
        if envs.VLLM_COMPUTE_NANS_IN_LOGITS:
            num_nans_in_logits = self._get_nans_in_logits(logits)

        # TODO(woosuk): The following loop can be slow since it iterates over
        # the requests one by one. Optimize.
        discard_sampled_tokens_req_indices = []
        for i, req_id in enumerate(self.input_batch.req_ids):
            req_state = self.requests[req_id]
            seq_len = (req_state.num_computed_tokens +
                       scheduler_output.num_scheduled_tokens[req_id])
            if seq_len < req_state.num_tokens:
                # Ignore the sampled token for partial prefills.
                # Rewind the generator state as if the token was not sampled.
                # This relies on cuda-specific torch-internal impl details
                generator = self.input_batch.generators.get(i)
                if generator is not None:
                    generator.set_offset(generator.get_offset() - 4)
                # Record the index of the request that should not be sampled,
                # so that we could clear the sampled tokens before returning.
                discard_sampled_tokens_req_indices.append(i)

        # NOTE: GPU -> CPU Sync happens here.
        # Move as many CPU operations as possible before this sync point.
        logprobs_tensors = sampler_output.logprobs_tensors
        logprobs_lists = logprobs_tensors.tolists() \
            if logprobs_tensors is not None else None

        # Compute prompt logprobs if needed.
        prompt_logprobs_dict = self._get_prompt_logprobs_dict(
            hidden_states[:num_scheduled_tokens],
            scheduler_output,
        )

        # Get the valid generated tokens.
        sampled_token_ids = sampler_output.sampled_token_ids
        max_gen_len = sampled_token_ids.shape[-1]
        if max_gen_len == 1:
            # No spec decode tokens.
            valid_sampled_token_ids = sampled_token_ids.tolist()
        else:
            # Includes spec decode tokens.
            valid_sampled_token_ids = self.rejection_sampler.parse_output(
                sampled_token_ids,
                self.input_batch.vocab_size,
            )
        # Mask out the sampled tokens that should not be sampled.
        for i in discard_sampled_tokens_req_indices:
            valid_sampled_token_ids[i].clear()

        # Cache the sampled tokens in the model runner, so that the scheduler
        # doesn't need to send them back.
        # NOTE(woosuk): As an exception, when using PP, the scheduler sends
        # the sampled tokens back, because there's no direct communication
        # between the first-stage worker and the last-stage worker.
        for req_idx, sampled_ids in enumerate(valid_sampled_token_ids):
            if not sampled_ids:
                continue

            start_idx = self.input_batch.num_tokens_no_spec[req_idx]
            end_idx = start_idx + len(sampled_ids)
            assert end_idx <= self.max_model_len, (
                "Sampled token IDs exceed the max model length. "
                f"Total number of tokens: {end_idx} > max_model_len: "
                f"{self.max_model_len}")

            self.input_batch.token_ids_cpu[req_idx,
                                           start_idx:end_idx] = sampled_ids
            self.input_batch.num_tokens_no_spec[req_idx] = end_idx
            self.input_batch.num_tokens[req_idx] = end_idx
            req_id = self.input_batch.req_ids[req_idx]
            req_state = self.requests[req_id]
            req_state.output_token_ids.extend(sampled_ids)

        if not self.speculative_config:
            # Speculative decoding is not enabled.
            spec_token_ids = None
        else:
            spec_token_ids = self.propose_draft_token_ids(
                scheduler_output,
                valid_sampled_token_ids,
                sampling_metadata,
                hidden_states,
                sample_hidden_states,
                aux_hidden_states,
                spec_decode_metadata,
                attn_metadata,
            )

        # Clear KVConnector state after all KVs are generated.
        if has_kv_transfer_group():
            get_kv_transfer_group().clear_connector_metadata()

        self.eplb_step()

        return ModelRunnerOutput(
            req_ids=self.input_batch.req_ids,
            req_id_to_index=self.input_batch.req_id_to_index,
            sampled_token_ids=valid_sampled_token_ids,
            spec_token_ids=spec_token_ids,
            logprobs=logprobs_lists,
            prompt_logprobs_dict=prompt_logprobs_dict,
            pooler_output=[],
            finished_sending=finished_sending,
            finished_recving=finished_recving,
            num_nans_in_logits=num_nans_in_logits,
        )




from __future__ import annotations

import inspect
import os
import warnings
from functools import wraps
from types import BuiltinFunctionType, BuiltinMethodType, FunctionType, MethodType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Union

import torch
from torch.nn.modules.module import _addindent

from .. import CONFIG, base_deprecation_message, deprecated, util
from ..util import apply, Patch

from .batching import Batchable
from .inject import convert as inject
from .tracing.base import Tracer, WithBlockNotFoundError
from .tracing.editing import EditingTracer
from .tracing.globals import Object
from .tracing.iterator import IteratorProxy
from .tracing.tracer import InterleavingTracer, ScanningTracer
from .interleaver import Interleaver



def trace_only(fn: Callable):

    @wraps(fn)
    def wrapper(self: Envoy, *args, **kwargs):

        if self._interleaver is None:
            raise ValueError(f"Must be within a trace to use `.{fn.__name__}(...)`")

        return fn(self, *args, **kwargs)

    return wrapper


class Envoy(Batchable):
    """
    A proxy class that wraps a PyTorch module to enable intervention during execution.

    This class provides access to module inputs and outputs during forward passes,
    and allows for modification of these values through an interleaving mechanism.
    It serves as the primary interface for inspecting and modifying the behavior
    of neural network modules during execution.

    Attributes:
        path (str): The module's location in the model hierarchy.
            Example: "model.encoder.layer1" indicates this module is the first layer of the encoder in the model.
        _module (torch.nn.Module): The underlying PyTorch module
        _source (Optional[EnvoySource]): Source code representation of the module
        _interleaver (Optional[Interleaver]): Interleaver for managing execution flow
        _default_mediators (List[List[str]]): List of default mediators created with .edit
        _children (List[Envoy]): List of child Envoys
        _alias (Aliaser): Aliaser object for managing aliases
    """

    def __init__(
        self,
        module: torch.nn.Module,
        interleaver: Optional[Interleaver] = None,
        path: Optional[str] = "model",
        rename: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> None:
        """
        Initialize an Envoy for a PyTorch module.

        Args:
            module (torch.nn.Module): The PyTorch module to wrap
            interleaver (Optional[Interleaver]): Optional interleaver for managing execution flow
            path (Optional[str]): Optional path string representing the module's location in the model hierarchy
            rename (Optional[Dict[str, Union[str, List[str]]]]): Optional dictionary mapping module names to alias names.
                Example: {"layer1": "first_layer", "layer2": "second_layer"}
                Example: {".model.layers": ".layers"} <-- Mounts .layers to the root model.
                Example: {".transformer": ["model", "mdl"]} <-- Allows access of .transformer as .model or .mdl

        """
        self.path = path

        self._module = module
        self._module.__path__ = path

        self._source = None

        self._interleaver = interleaver if interleaver is not None else Interleaver()
        self._interleaver.wrap_module(module)

        self._default_mediators: List[List[str]] = []

        self._children: List[Envoy] = []

        self._fake_inputs = inspect._empty
        self._fake_output = inspect._empty

        if rename is not None:
            self._alias = Aliaser(rename)
        else:
            self._alias = None

        for name, module in list(self._module.named_children()):
            setattr(self, name, module)

        if rename is not None:
            self._alias.build(self)

    def __getitem__(self, key: str) -> Envoy:
        """
        Access a child Envoy by index for Module Lists.

        Args:
            key: The index of the child Envoy to retrieve

        Returns:
            The child Envoy at the specified index
        """
        return self._children[key]

    @property
    def interleaving(self) -> bool:
        """
        Check if the Envoy is currently nterleaving.

        Returns:
            True if the Envoy is interleaving, False otherwise
        """
        return self._interleaver is not None and self._interleaver.interleaving

    #### Properties ####

    @property
    def output(self) -> Object:
        """
        Get the output of the module's forward pass.

        This property allows access to the return values produced by the module
        during the forward pass.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     attn = model.transformer.h[0].attn.output[0].save()
            >>> print(attn)

        Returns:
            The module's output values
        """

        if self.interleaving:

            return self._interleaver.current.request(
                self._interleaver.current.iterate(f"{self.path}.output")
            )
        elif self._fake_output is not inspect._empty:
            return self._fake_output
        else:
            raise ValueError(
                "Cannot return output of Envoy that is not interleaving nor has a fake output set."
            )

    @output.setter
    def output(self, value: Any):
        """
        Set new values for the module's output.

        This allows for intervention by replacing the module's output with
        custom values during execution.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     model.transformer.h[0].attn.output[0] *= 2

        Args:
            value: The new output value to use.
        """
        if self.interleaving:
            self._interleaver.current.swap(
                self._interleaver.current.iterate(f"{self.path}.output"), value
            )

        else:
            raise ValueError("Cannot set output of Envoy that is not interleaving.")

    @property
    def inputs(self) -> Tuple[Tuple[Object], Dict[str, Object]]:
        """
        Get the inputs to the module's forward pass.

        This property provides access to all input values passed to the module
        during the forward pass.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     args, kwargs = model.transformer.h[0].attn.inputs

        Returns:
            The module's input values as a tuple of positional and keyword arguments. i.e (args, kwargs)

        """

        if self.interleaving:
            return self._interleaver.current.request(
                self._interleaver.current.iterate(f"{self.path}.input")
            )
        elif self._fake_inputs is not inspect._empty:
            return self._fake_inputs
        else:
            raise ValueError(
                "Cannot return inputs of Envoy that is not interleaving nor has a fake inputs set."
            )

    @inputs.setter
    def inputs(self, value: Any):
        """
        Set new values for the module's inputs.

        This allows for intervention by replacing the module's inputs with
        custom values during execution.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     model.transformer.h[0].attn.inputs = (torch.randn(1, 1024, 1024), {})

        Args:
            value: The new input value(s) to use, structured as a tuple of (args, kwargs)
        """
        if self.interleaving:
            self._interleaver.current.swap(
                self._interleaver.current.iterate(f"{self.path}.input"), value
            )
        else:
            raise ValueError("Cannot set inputs of Envoy that is not interleaving.")

    @property
    def input(self) -> Object:
        """
        Get the first input to the module's forward pass.

        This is a convenience property that returns just the first input value
        from all inputs passed to the module. So first positional argument, or first keyword argumetn if there are no positional arguments.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     hidden_states = model.transformer.h[0].attn.input.save()
            >>> print(hidden_states)

        Returns:
            The first input value
        """

        inputs = self.inputs

        return [*inputs[0], *inputs[1].values()][0]

    @input.setter
    def input(self, value: Any):
        """
        Set a new value for the module's first input.

        This is a convenience method that replaces just the first input value
        while preserving all other inputs.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     model.transformer.h[0].attn.input = torch.randn(1, 1024, 1024)

        Args:
            value: The new value for the first input
        """
        inputs = self.inputs

        value = (value, *inputs[0][1:]), inputs[1]

        self.inputs = value

    @property
    def source(self) -> EnvoySource:
        """
        Get the source code representation of the module.

        This property provides access to the module's source code with operations
        highlighted, allowing for inspection and intervention at specific points.

        Example:

            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)

            >>> # We can print to see the formward method of the module and names associated with the operations within.
            >>> print(model.transformer.h[0].attn.source)

                                                   60
                                                   61     if using_eager and self.reorder_and_upcast_attn:
              self__upcast_and_reordered_attn_0 -> 62         attn_output, attn_weights = self._upcast_and_reordered_attn(
                                                   63             query_states, key_states, value_states, attention_mask, head_mask
                                                   64         )
                                                   65     else:
              attention_interface_0             -> 66         attn_output, attn_weights = attention_interface(
                                                   67             self,
                                                   68             query_states,
                                                   69             key_states,
                                                   70             value_states,
                                                   71             attention_mask,
                                                   72             head_mask=head_mask,
                                                   73             dropout=self.attn_dropout.p if self.training else 0.0,
                                                   74             is_causal=is_causal,
                                                   75             **kwargs,
                                                   76         )
                                                   77
              attn_output_reshape_0             -> 78     attn_output = attn_output.reshape(*attn_output.shape[:-2], -1).contiguous()
              contiguous_0                      ->  +     ...
              self_c_proj_0                     -> 79     attn_output = self.c_proj(attn_output)
              self_resid_dropout_0              -> 80     attn_output = self.resid_dropout(attn_output)
                                                   81
                                                   82     return attn_output, attn_weights
                                                   83

            >>> # We can print out one of these to see the only the operation and a few operations before and after.
            >>> print(model.transformer.h[0].attn.source.attention_interface_0)

            .transformer.h.0.attn.attention_interface_0:

                 ....

                     if using_eager and self.reorder_and_upcast_attn:
                         attn_output, attn_weights = self._upcast_and_reordered_attn(
                             query_states, key_states, value_states, attention_mask, head_mask
                         )
                     else:
                 -->     attn_output, attn_weights = attention_interface( <--
                             self,
                             query_states,
                             key_states,
                             value_states,
                             attention_mask,
                             head_mask=head_mask,
                 ....

            >>> with model.trace("Hello World"):
            >>>     # Now we can access it like we would any other Envoy with .input or .output to grab the intermediate value.
            >>>     attn = model.transformer.h[0].attn.source.attention_interface_0.output.save()

            >>> print(attn)


        Returns:
            An EnvoySource object containing the module's source code and operations
        """
        if self._source is None:

            def wrap(fn: Callable, **kwargs):

                bound_obj = (
                    fn.__self__
                    if inspect.ismethod(fn) and fn.__name__ != "forward"
                    else None
                )

                if self.interleaving:
                    return self._interleaver.wrap_operation(
                        fn, **kwargs, bound_obj=bound_obj
                    )
                else:
                    return fn

            source, line_numbers, forward = inject(
                self._module.forward, wrap, self._module.__path__
            )
            self._module.forward = MethodType(forward, self._module)

            self._source = EnvoySource(self._module.__path__, source, line_numbers, interleaver=self._interleaver)

        return self._source

    def __call__(self, *args, hook: bool = False, **kwargs):
        return (
            self._module.forward(*args, **kwargs)
            if self.interleaving and not hook
            else self._module(*args, **kwargs)
        )

    #### Public methods ####

    def trace(self, *args, fn: Optional[Callable] = None, trace: bool = None, tracer_cls: Type[InterleavingTracer] = InterleavingTracer, **kwargs):
        """
        Create a tracer for this module.

        This method returns a tracer that can be used to capture and modify
        the execution of the module.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     model.transformer.h[0].attn.output[0][:] = 0

            >>>     output = model.output.save()
            >>> print(output)

        Args:
            *args: Arguments to pass to the tracer
            **kwargs: Keyword arguments to pass to the tracer

        Returns:
            An InterleavingTracer for this module
        """

        # TODO trace= is Legacy
        if trace is not None:
            deprecation_message = f"The `trace` argument {base_deprecation_message}\nJust call the method without a with context instead."
            warnings.warn(deprecation_message)

        if fn is None:
            fn = self.__call__
            kwargs["hook"] = True

        return tracer_cls(fn, self, *args, **kwargs)

    def scan(self, *args, **kwargs):
        """
        Just like .trace() but runs the model in fake tensor mode to validate operations and inspect tensor shapes.

        This method returns a tracer that runs the model in fake tensor mode to validate operations
        and inspect tensor shapes without performing actual computation. This is useful for:
        - Validating that operations will work with given input shapes
        - Inspecting the shapes and types of tensors that would flow through the model
        - Debugging shape mismatches or other tensor-related issues.

        Note this will not dispatch the model if not dispatched.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> # Value error as the fake inputs and outputs have not been scanned in.
            >>> print(model.transformer.h[0].mlp.output.shape)
            >>> # Scan the model to validate operations and inspect shapes
            >>> with model.scan("Hello World"):
            >>>     # Access fake inputs/outputs to inspect shapes
            >>>     attn_input = model.transformer.h[0].attn.input.save()
            >>>     attn_output = model.transformer.h[0].attn.output[0].save()
            >>> print(f"Attention input shape: {attn_input.shape}")
            >>> print(f"Attention output shape: {attn_output.shape}")
            >>> print(model.transformer.h[0].mlp.output.shape)

        Args:
            *args: Arguments to pass to the tracer
            **kwargs: Keyword arguments to pass to the tracer

        Returns:
            A ScanningTracer for this module
        """
        return ScanningTracer(self.__call__, self, *args, hook=True, **kwargs)

    def edit(self, *, inplace: bool = False):
        """
        Create an editing tracer for this module. Allows for setting default interventions.
        This means this tracer won't execute the module, but will instead set default interventions that are applied on all future executions.

        Edits can be cleared with `Envoy.clear_edits()`.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> # Now the first layer attention output will always be 0.
            >>> with model.edit() as edited_model:
            >>>     edited_model.transformer.h[0].attn.output[:] = 0


            >>> with model.trace("Hello World"):
            >>>     output = model.output.save()
            >>> # The orignal model will have the default output.
            >>> print(output)

            >>> with edited_model.trace("Hello World"):
            >>>     edited_output = edited_model.output.save()
            >>> # The edited model will have the output after our intervention.
            >>> print(edited_output)


        Args:
            inplace (bool, optional): Whether to edit in place. Defaults to False.

        Returns:
            (EditingTracer): An EditingTracer for this module
        """

        return EditingTracer(self.__call__, self, inplace=inplace)

    def clear_edits(self):
        """
        Clear all edits for this Envoy.
        """
        self._default_mediators = []

    def export_edits(
        self, name: str, export_dir: Optional[str] = None, variant: str = "__default__"
    ):
        """TODO

        Args:
            name (str): _description_
            export_dir (Optional[str], optional): _description_. Defaults to None.
            variant (str, optional): _description_. Defaults to '__default__'.

        Raises:
            ValueError: _description_
        """

        if len(self._default_mediators) == 0:
            raise ValueError("Cannot export an Envoy before calling .edit().")

        if export_dir is None:

            export_dir = os.path.join(CONFIG.APP.CACHE_DIR, "exports")

        export_dir = os.path.expanduser(os.path.join(export_dir, name))

        os.makedirs(export_dir, exist_ok=True)
        
        from . import serialization

        serialization.save(
            self._default_mediators, os.path.join(export_dir, f"{variant}.dill")
        )

    def import_edits(
        self, name: str, export_dir: Optional[str] = None, variant: str = "__default__"
    ):
        """TODO

        Args:
            name (str): _description_
            export_dir (Optional[str], optional): _description_. Defaults to None.
            variant (str, optional): _description_. Defaults to '__default__'.
        """

        if export_dir is None:

            export_dir = os.path.join(CONFIG.APP.CACHE_DIR, "exports")

        export_dir = os.path.expanduser(os.path.join(export_dir, name))
        
        from . import serialization

        imported_mediators = serialization.load(
            os.path.join(export_dir, f"{variant}.dill"), self
        )

        self._default_mediators.extend(imported_mediators)

    # TODO legacy
    def session(self, *args, tracer_cls: Type[Tracer] = Tracer, **kwargs):
        tracer = tracer_cls(*args, **kwargs)
        setattr(tracer, "model", self)
        return tracer

    @property
    @deprecated(message="Use `tracer.iter` instead.")
    @trace_only
    def iter(self):
        return IteratorProxy(self._interleaver)

    @deprecated(message="Use `tracer.all()` instead.")
    @trace_only
    def all(self):
        return self.iter[:]
    
    @deprecated(message="Use `tracer.next()` instead.")
    @trace_only
    def next(self, step: int = 1):
        self._interleaver.current.iteration += step
        
        return self

    @trace_only
    def skip(self, replacement: Any):
        """Skips the execution of this module duting execution / interleaving.
        Behavior is the module will not be executed and will return a replacement value instead.

        Example:
            >>> model = LanguageModel("gpt2", device_map='auto', dispatch=True)
            >>> with model.trace("Hello World"):
            >>>     # Skip the first layer and replace it with the input to the layer.
            >>>     model.transformer.h[0].skip((model.transformer.h[0].input, None))
            >>>     output = model.output.save()
            >>> print(output)

        Args:
            replacement (Any): The replacement value to replace the module's output with.
        """

        requester = self._interleaver.current.iterate(f"{self.path}.input")

        self._interleaver.current.skip(requester, replacement)

    @trace_only
    def wait_for_input(self):
        """
        Wait for the input to the module to be available.
        """
        self.inputs

    @trace_only
    def wait_for_output(self):
        """
        Wait for the output to the module to be available.
        """
        self.output

    def to(self, device: torch.device):
        """
        Move the module to a specific device.

        This method moves the underlying PyTorch module to the specified device.

        Args:
            device: The device to move the module to

        Returns:
            Self, for method chaining
        """
        self._module.to(device)

        return self

    def cpu(self, *args, **kwargs):
        """
        Move the module to the CPU.
        """
        self._module.cpu(*args, **kwargs)
        return self

    def cuda(self, *args, **kwargs):
        """
        Move the module to the GPU.
        """
        self._module.cuda(*args, **kwargs)
        return self

    @property
    def device(self) -> Optional[torch.device]:
        """
        Get the device the module is on. Finds the first parameter and return its device.
        """
        try:
            return next(self._module.parameters()).device
        except:
            return None

    def modules(
        self,
        include_fn: Callable[[Envoy], bool] = None,
        names: bool = False,
    ) -> List[Envoy]:
        """
        Get all modules in the Envoy tree.

        This method returns all Envoys in the tree, optionally filtered by
        an inclusion function.

        Args:
            include_fn: Optional function to filter modules
            names: Whether to include module names in the result

        Returns:
            A list of Envoys or (name, Envoy) tuples
        """
        result = []

        for envoy in self._children:

            result.extend(envoy.modules(include_fn=include_fn, names=names))

        if include_fn is None or include_fn(self):

            if names:
                result.append((self.path, self))
            else:
                result.append(self)

        return result

    def named_modules(self, *args, **kwargs) -> List[Tuple[str, Envoy]]:
        """
        Returns all Envoys in the Envoy tree along with their name/module_path.

        This is a convenience method that calls modules() with names=True.

        Args:
            include_fn (Callable, optional): Optional function to be ran against all Envoys to check if they should be included in the final collection of Envoys. Defaults to None.
            *args, **kwargs: Additional arguments to pass to modules()

        Returns:
            List[Tuple[str, Envoy]]: Included Envoys and their names/module_paths.
        """

        return self.modules(*args, **kwargs, names=True)

    def get(self, path: str) -> Object:
        """Gets the Envoy/Proxy via its path.

        e.x:
            model = nnsight.LanguageModel("openai-community/gpt2")

            module = model.get('transformer.h.0.mlp')

            with model.trace("Hello"):
                value = model.get('transformer.h.0.mlp.output').save()

        Args:
            path (str): '.' separated path.

        Returns:
            Union[Envoy, InterventionProxyType]: Fetched Envoy/Proxy
        """
        return util.fetch_attr(self, path)

    def interleave(self,  fn: Union[Callable, str], *args, **kwargs):

        device = self.device

        (args, kwargs) = apply(
            (args, kwargs), lambda tensor: tensor.to(device), torch.Tensor
        )
        
        if isinstance(fn, str):
            fn = getattr(self, fn)

        try:
            with self._interleaver:
                fn(*args, **kwargs)
                
            self._interleaver.check_cache_full()
            self._interleaver.check_dangling_mediators()
            
        finally:
            self._interleaver.cancel()

    #### Private methods ####

    def _add_envoy(self, module: torch.nn.Module, name: str) -> Envoy:
        """
        Adds a new Envoy for a given torch module under this Envoy.

        This method creates a new Envoy for a child module and adds it to
        this Envoy's children.

        Args:
            module: Module to create Envoy for.
            name: Name of envoy/attribute.
        """
        module_path = f"{self.path}.{name}"

        envoy = Envoy(
            module,
            path=module_path,
            rename=self._alias.rename if self._alias is not None else None,
            interleaver=self._interleaver
        )

        self._children.append(envoy)

        setattr(self._module, name, module)

        # If the module already has a sub-module named 'input' or 'output',
        # mount the proxy access to 'nns_input' or 'nns_output instead.

        if hasattr(Envoy, name):
            self._handle_overloaded_mount(envoy, name)
        else:
            super().__setattr__(name, envoy)

        return envoy

    def _handle_overloaded_mount(self, envoy: Envoy, mount_point: str) -> None:
        """If a given module already has an attribute of the same name as something nnsight wants to add, we need to rename it.

        Directly edits the underlying class to accomplish this.

        Args:
            envoy (Envoy): Envoy to handle.
            mount_point (str): Overloaded attribute name.
        """

        warnings.warn(
            f"Module `{self.path}` of type `{type(self._module)}` has pre-defined a `{mount_point}` attribute. nnsight access for `{mount_point}` will be mounted at `.nns_{mount_point}` instead of `.{mount_point}` for this module only."
        )

        # If we already shifted a mount point dont create another new class.
        if "Preserved" in self.__class__.__name__:

            new_cls = self.__class__

        else:

            new_cls = type(
                f"{self.__class__.__name__}.Preserved",
                (self.__class__,),
                {},
            )

            self.__class__ = new_cls

        # Get the normal proxy mount point
        mount = getattr(Envoy, mount_point)

        setattr(new_cls, f"nns_{mount_point}", mount)

        if isinstance(mount, property):

            mount = property(
                lambda slf: slf.__dict__[mount_point],
                mount.fset,
                mount.fdel,
                mount.__doc__,
            )

            setattr(new_cls, mount_point, mount)

        # Move it to nns_<mount point>
        self.__dict__[mount_point] = envoy

    def _update(self, module: torch.nn.Module) -> None:
        """Updates the ._model attribute using a new model of the same architecture.
        Used when loading the real weights (dispatching) and need to replace the underlying modules.
        """

        i = 0

        for i, child in enumerate(module.children()):

            self._children[i]._update(child)

        # Handle extra modules added after initialization: issues/376
        for name, child in list(self._module.named_children())[i + 1 :]:

            setattr(module, name, child)

        self._module = module
        self._module.__path__ = self.path
        self._interleaver.wrap_module(module)

        if self._source is not None:

            def wrap(fn: Callable, **kwargs):

                bound_obj = (
                    fn.__self__
                    if inspect.ismethod(fn) and fn.__name__ != "forward"
                    else None
                )

                if self.interleaving:
                    return self._interleaver.wrap_operation(
                        fn, **kwargs, bound_obj=bound_obj
                    )
                else:
                    return fn

            source, line_numbers, forward = inject(
                self._module.forward, wrap, self._module.__path__
            )
            self._module.forward = MethodType(forward, self._module)

    def _update_alias(self, alias: Dict[str, str]):
        """
        Update the alias for this Envoy and its children.
        """
        if alias is not None:
            self._alias = Aliaser(alias)
            self._alias.build(self)

            for envoy in self._children:
                envoy._update_alias(alias)



    def _shallow_copy(self) -> Envoy:
        """Creates a new instance copy of the same class with the all the attributes of the original instance.

        Returns:
            Self: NNsightModel
        """
        copy = self.__class__.__new__(self.__class__)
        for key, value in self.__dict__.items():
            copy.__dict__[key] = value

        return copy

    #### Dunder methods ####

    def __len__(self):
        """
        Get the length of the Envoy.
        """
        return len(self._module)

    def __iter__(self):
        """
        Iterate over the Envoy.
        """
        return iter(self._children)

    def __str__(self):
        """
        String representation of the Envoy.

        Returns:
            A string representation of the Envoy showing its path
        """
        return self.__repr__()

    def __reprlist__(self):

        list_of_reprs = [repr(item) for item in self]
        if len(list_of_reprs) == 0:
            return self._module._get_name() + "()"

        start_end_indices = [[0, 0]]
        repeated_blocks = [list_of_reprs[0]]
        for i, r in enumerate(list_of_reprs[1:], 1):
            if r == repeated_blocks[-1]:
                start_end_indices[-1][1] += 1
                continue

            start_end_indices.append([i, i])
            repeated_blocks.append(r)

        lines = []
        main_str = self._module._get_name() + "("
        for (start_id, end_id), b in zip(start_end_indices, repeated_blocks):
            local_repr = f"({start_id}): {b}"  # default repr

            if start_id != end_id:
                n = end_id - start_id + 1
                local_repr = f"({start_id}-{end_id}): {n} x {b}"

            local_repr = _addindent(local_repr, 2)
            lines.append(local_repr)

        main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str

    def __repr__(self):
        """
        Representation of the Envoy.

        Returns:
            The string representation of the Envoy
        """

        if isinstance(self._module, torch.nn.ModuleList):
            return self.__reprlist__()

        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = self._module.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split("\n")
        child_lines = []
        for envoy in self._children:
            key = envoy.path.split(".")[-1]
            mod_str = repr(envoy)
            mod_str = _addindent(mod_str, 2)
            if self._alias is not None and key in self._alias.name_to_aliases:
                key = "/".join([*self._alias.name_to_aliases[key], key])
            child_lines.append("(" + key + "): " + mod_str)

        if self._alias is not None:
            for extra in self._alias.extras:

                key = "/".join(self._alias.name_to_aliases[extra])
                envoy = self.get(extra)
                mod_str = repr(envoy)
                mod_str = _addindent(mod_str, 2)
                child_lines.append("(" + key + "): " + mod_str)

        lines = extra_lines + child_lines

        main_str = self._module._get_name() + "("
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"
        return main_str

    def __getattr__(self, name: str) -> Union[torch.nn.Module, Envoy, Any]:
        """
        Get an attribute from the underlying module.

        If the attribute is callable, it will be wrapped in a tracer to enable
        intervention during execution.

        Args:
            name: The name of the attribute to get

        Returns:
            The attribute value, possibly wrapped in a tracer

        Raises:
            AttributeError: If the attribute doesn't exist
        """

        if self._alias is not None and name in self._alias.alias_to_name:
            return util.fetch_attr(self, self._alias.alias_to_name[name])

        if hasattr(self._module, name):
            value = getattr(self._module, name)

            # It's a method bound to the module, create an interleaver for it
            if isinstance(
                value,
                (FunctionType, MethodType, BuiltinFunctionType, BuiltinMethodType),
            ):

                # If the Envoy defines a method with __nnsight_{name}__, use it instead to override
                value = getattr(self, f"__nnsight_{name}__", value)

                def trace(*args, **kwargs):
                    try:
                        return self.trace(*args, fn=value, **kwargs)

                    except WithBlockNotFoundError as e:

                        return value(*args, **kwargs)

                return trace

            elif isinstance(value, torch.nn.Module):
                # If the _module has a module in its __dict__ but wasn't picked up when creating the Envoy,
                # Hopefully it is alrady an Envoy somewhere in the tree.
                # https://github.com/ndif-team/nnsight/issues/479
                # This happened because some transformers models set this class attr: _checkpoint_conversion_mapping
                if hasattr(value, "__path__"):
                    return util.fetch_attr(self, value.__path__[len(self.path) :])
                return self._add_envoy(value, name)
            else:
                return value
        else:
            raise AttributeError(f"{self} has no attribute {name}")

    def __setattr__(self, key: Any, value: Any) -> None:
        """
        Set an attribute on the Envoy.

        If the value is a PyTorch module, it will be wrapped in an Envoy to enable
        intervention during execution.

        Args:
            key: The attribute name
            value: The attribute value
        """
        if key != "_module" and isinstance(value, torch.nn.Module):
            self._add_envoy(value, key)
        else:
            super().__setattr__(key, value)


# TODO extend Envoy
class OperationEnvoy:
    """
    Represents a specific operation within a module's forward pass.

    This class provides access to the inputs and outputs of individual
    operations within a module's execution, allowing for fine-grained
    inspection and intervention at the operation level.
    """

    def __init__(
        self,
        name: str,
        source: str,
        line_number: int,
        interleaver: Optional[Interleaver] = None,
    ):
        """
        Initialize an OperationEnvoy.

        Args:
            name: The fully qualified name of the operation
            source: The source code of the module containing the operation
            line_number: The line number of the operation in the source
            interleaver: Optional interleaver for managing execution flow
        """
        self.name = name
        self.source_code = source
        self.line_number = line_number

        self._interleaver = interleaver

        self._source: EnvoySource = None
        self._fn: Callable = None

    def __str__(self):
        """
        String representation showing the operation in context.

        This method returns a formatted string showing the operation's source code
        with surrounding context lines and highlighting the operation line.

        Returns:
            A formatted string showing the operation's source code with context
        """
        source_lines = self.source_code.split("\n")
        start_idx = max(0, self.line_number - 5)
        end_idx = min(len(source_lines) - 1, self.line_number + 8)

        highlighted_lines = [self.name + ":\n"]

        if start_idx != 0:
            highlighted_lines.append("    ....")

        for i in range(start_idx, end_idx):
            line = source_lines[i]
            if i == self.line_number + 1:
                highlighted_lines.append(f"    --> {line[4:]} <--")
            else:
                highlighted_lines.append("    " + line)

        if end_idx != len(source_lines) - 1:
            highlighted_lines.append("    ....")

        return "\n".join(highlighted_lines)

    @property
    def output(self) -> Union[Any, torch.Tensor]:
        """
        Get the output of this operation.

        This property provides access to the return value(s) produced by the operation
        during execution.

        Returns:
            The operation's output value(s)
        """

        return self._interleaver.current.request(f"{self.name}.output")

    @output.setter
    def output(self, value: Any) -> None:
        """
        Set a new value for the operation's output.

        This allows for intervention by replacing the operation's output with
        a custom value during execution.

        Args:
            value: The new output value
        """
        self._interleaver.current.swap(f"{self.name}.output", value)

    @property
    def inputs(
        self,
    ) -> Tuple[Tuple[Any, torch.Tensor], Dict[str, Union[torch.Tensor, Any]]]:
        """
        Get the inputs to this operation.

        This property provides access to all input value(s) passed to the operation
        during execution, structured as a tuple of positional and keyword arguments.

        Returns:
            The operation's input value(s)
        """
        return self._interleaver.current.request(f"{self.name}.input")

    @inputs.setter
    def inputs(self, value: Any) -> None:
        """
        Set new values for the operation's inputs.

        This allows for intervention by replacing the operation's inputs with
        custom values during execution.

        Args:
            value: The new input value(s)
        """
        self._interleaver.current.swap(f"{self.name}.input", value)

    @inputs.deleter
    def inputs(self):
        """
        Clear the cached input value.

        This removes any stored input values, forcing them to be recomputed
        on the next access.
        """
        self._input = None

    @property
    def input(self) -> Union[Any, torch.Tensor]:
        """
        Get the first input to the operation.

        This is a convenience property that returns just the first input value
        from all inputs passed to the operation.

        Returns:
            The first input value
        """

        inputs = self.inputs

        return [*inputs[0], *inputs[1].values()][0]

    @input.setter
    def input(self, value: Any) -> None:
        """
        Set a new value for the operation's first input.

        This is a convenience method that replaces just the first positional input
        while preserving all other inputs.

        Args:
            value: The new value for the first input
        """
        inputs = self.inputs

        value = (value, *inputs[0][1:]), inputs[1]

        self.inputs = value

    @property
    def source(self) -> EnvoySource:
        """
        Get the source code of the operation.

        This property provides access to the operation's source code with nested
        operations highlighted, allowing for inspection and intervention at specific points.

        Returns:
            An EnvoySource object containing the operation's source code and nested operations
        """

        if self._source is None:
            fn = self._interleaver.current.request(f"{self.name}.fn")

            # TODO maybe do something else here
            if isinstance(fn, torch.nn.Module):

                msg = f"Don't call .source on a module ({getattr(fn, '__path__', '')}) from within another .source. Call it directly with: {getattr(fn, '__path__', '')}.source"
                raise ValueError(msg)

            def wrap(fn: Callable, **kwargs):

                bound_obj = (
                    fn.__self__
                    if getattr(fn, "__name__", None) != "forward"
                    and inspect.ismethod(fn)
                    else None
                )

                return self._interleaver.wrap_operation(
                    fn, **kwargs, bound_obj=bound_obj
                )

            source, line_numbers, fn = inject(fn, wrap, self.name)

            self._source = EnvoySource(
                self.name, source, line_numbers, interleaver=self._interleaver
            )

            self._fn = fn

        if f"{self.name}.fn" not in self._interleaver.current.history:
            self._interleaver.current.swap(f"{self.name}.fn", self._fn)

        return self._source

    # @input.setter
    # def input(self, value: Any):
    #     #TODO would need await...
    #     inputs = self._input
    #     self._input = ((value, *inputs[0]), inputs[1])
    #     self.interleaver.set_swap(self._input, (self.module, self.name), Events.INPUT)

    # @input.deleter
    # def input(self):
    #     self._input = None


class EnvoySource:
    """
    Represents the source code of a module with operations highlighted.

    This class provides access to the individual operations within a module's
    source code, allowing for inspection and intervention at specific points
    in the code. It serves as a bridge between the source code representation
    and the runtime execution of operations.
    """

    def __init__(
        self,
        name: str,
        source: str,
        line_numbers: dict,
        interleaver: Optional[Interleaver] = None,
    ):
        """
        Initialize an EnvoySource.

        Args:
            name: The fully qualified name of the module or operation
            source: The source code string
            line_numbers: A dictionary mapping operation names to line numbers
            interleaver: Optional interleaver for managing execution flow
        """
        self.source = source
        self.line_numbers = line_numbers

        self.operations: List[OperationEnvoy] = []

        for _name, line_number in line_numbers.items():
            operation = OperationEnvoy(
                f"{name}.{_name}", source, line_number, interleaver=interleaver
            )
            setattr(self, _name, operation)
            self.operations.append(operation)

    def __str__(self):
        """
        String representation showing the source code with operations highlighted.

        This method returns a formatted string showing the source code with
        operation names and line numbers, making it easy to identify intervention points.

        Returns:
            A formatted string showing the source code with operation names and line numbers
        """
        # Find the longest name for proper alignment
        max_name_length = (
            max(len(name) for name in self.line_numbers.keys())
            if self.line_numbers
            else 0
        )

        source_lines = self.source.split("\n")
        formatted_lines = [
            " " * (max_name_length + 6) + "* " + source_lines[0]
        ]  # Keep the function definition unchanged

        # Group operations by line number
        operations_by_line = {}
        for name, line_number in self.line_numbers.items():
            if line_number not in operations_by_line:
                operations_by_line[line_number] = []
            operations_by_line[line_number].append(name)

        for i, line in enumerate(source_lines[1:]):
            line_number = i

            # Check if this line has operations
            if line_number in operations_by_line:
                # Handle multiple operations on the same line
                operations = operations_by_line[line_number]

                # First operation gets the line number
                first_op = operations[0]
                line_prefix = f" {first_op:{max_name_length}} ->{line_number:3d} "
                formatted_lines.append(f"{line_prefix}{line}")

                # For nested operations, unwrap them onto separate lines
                if len(operations) > 1:
                    for op in operations[1:]:
                        continuation_prefix = f" {op:{max_name_length}} ->  + "
                        # Instead of just showing a vertical line, show the operation on its own line
                        formatted_lines.append(
                            f"{continuation_prefix}{' ' * (len(line) - len(line.lstrip()))}..."
                        )
            else:
                # Regular line with no operations
                line_prefix = " " * (max_name_length + 4) + f"{line_number:3d} "
                formatted_lines.append(f"{line_prefix}{line}")

        source = "\n".join(formatted_lines)

        return source



    def __getattribute__(self, name: str) -> Union[OperationEnvoy]:

        return object.__getattribute__(self, name)


class Aliaser:

    def __init__(self, rename: Dict[str, Union[str, List[str]]]):
        """
        Initialize an Aliaser.

        Args:
            rename (Dict[str, Union[str, List[str]]]): Dictionary mapping module names to alias names.
                Example: {"layer1": "first_layer", "layer2": "second_layer"}
                Example: {".model.layers": ".layers"} <-- Mounts .layers to the root model.
                Example: {".transformer": ["model", "mdl"]} <-- Allows access of .transformer as .model or .mdl

        Attributes:
            rename (Dict[str, Union[str, List[str]]]): Dictionary mapping module names to alias names.
            alias_to_name (Dict[str, str]): Dictionary mapping alias names to module names.
            name_to_aliases (Dict[str, List[str]]): Dictionary mapping module names to list of alias names.
            extras (Dict[str, List[str]]): Dictionary mapping attribute paths (.transformer.h) to list of alias names.
                Used to show dot seperated attributes in the string representation of the Envoy.


        """

        self.rename = rename

        self.alias_to_name = {}
        self.name_to_aliases = {}
        self.extras = {}

    def build(self, envoy: Envoy):

        for name, aliases in self.rename.items():

            try:
                util.fetch_attr(envoy, name)
            except:
                continue

            if isinstance(aliases, str):
                aliases = [aliases]

            name = name.removeprefix(".")

            if "." in name:

                self.extras[name] = aliases

            self.name_to_aliases[name] = aliases

            for alias in aliases:
                self.alias_to_name[alias] = name


import ctypes
import inspect
from types import FrameType
from typing import Any, Callable, Dict, Union

from typing_extensions import Self

from ...intervention.backends import Backend
from ...intervention.backends.remote import RemoteBackend
from ...intervention.serialization import load, save
from ...intervention.tracing.tracer import InterleavingTracer, Tracer
from ...util import from_import_path, to_import_path
from .meta import MetaMixin


class RemoteableMixin(MetaMixin):

    def trace(
        self,
        *inputs: Any,
        backend: Union[Backend, str, None] = None,
        remote: bool = False,
        blocking: bool = True,
        **kwargs: Dict[str, Any],
    ):

        if backend is not None:
            pass
        elif remote:
            backend = RemoteBackend(self.to_model_key(), blocking=blocking)
        # If backend is a string, assume RemoteBackend url.
        elif isinstance(backend, str):
            backend = RemoteBackend(
                self.to_model_key(), host=backend, blocking=blocking
            )
        return super().trace(
            *inputs,
            backend=backend,
            tracer_cls=RemoteInterleavingTracer,
            **kwargs,
        )

    def session(
        self,
        *inputs: Any,
        backend: Union[Backend, str, None] = None,
        remote: bool = False,
        blocking: bool = True,
        **kwargs: Dict[str, Any],
    ):

        if backend is not None:
            pass
        elif remote:
            backend = RemoteBackend(self.to_model_key(), blocking=blocking)
        # If backend is a string, assume RemoteBackend url.
        elif isinstance(backend, str):
            backend = RemoteBackend(
                self.to_model_key(), host=backend, blocking=blocking
            )
        return super().session(
            *inputs,
            backend=backend,
            tracer_cls=RemoteTracer,
            **kwargs,
        )

    def _remoteable_model_key(self) -> str:

        raise NotImplementedError()

    @classmethod
    def _remoteable_from_model_key(cls, model_key: str) -> Self:
        raise NotImplementedError()

    def to_model_key(self) -> str:

        return f"{to_import_path(type(self))}:{self._remoteable_model_key()}"

    @classmethod
    def from_model_key(cls, model_key: str, **kwargs) -> Self:

        import_path, model_key = model_key.split(":", 1)

        type: RemoteableMixin = from_import_path(import_path)

        return type._remoteable_from_model_key(model_key, **kwargs)


class StreamTracer(Tracer):

    _send: Callable = None
    _recv: Callable = None

    def __init__(self, frame: FrameType, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.frame = frame
        
    @classmethod
    def register(cls, send_fn: Callable, recv_fn: Callable):
        
        cls._send = send_fn
        cls._recv = recv_fn
        
    @classmethod
    def deregister(cls):
        
        cls._send = None
        cls._recv = None
        
    def execute(self, fn: Callable):

        data = save(fn)

        if self._send is None:
            raise ValueError("No send function provided")

        if self._recv is None:
            raise ValueError("No recv function provided")

        self._send(data)

        data: Dict[str, Any] = load(self._recv(), None)

        if self.frame.f_code.co_filename.startswith("<nnsight"):
            # For dynamically generated code, update both globals and locals
            self.frame.f_globals.update(data)
            self.frame.f_locals.update(data)

            # Ensure locals are properly synchronized with the frame
            ctypes.pythonapi.PyFrame_LocalsToFast(
                ctypes.py_object(self.frame), ctypes.c_int(0)
            )

        else:
            # For regular files, just update locals
            for key, value in data.items():

                self.frame.f_locals[key] = value

                ctypes.pythonapi.PyFrame_LocalsToFast(
                    ctypes.py_object(self.frame), ctypes.c_int(0)
                )
        
class RemoteTracer(Tracer):

    def local(self):
        
        frame = inspect.currentframe().f_back
        
        return StreamTracer(frame)


class RemoteInterleavingTracer(InterleavingTracer, RemoteTracer):
    pass



from typing import Callable, TYPE_CHECKING, Any

from ..interleaver import Mediator
from .base import Tracer
from .util import try_catch


if TYPE_CHECKING:
    from .tracer import InterleavingTracer
else:
    InterleavingTracer = Any


class Invoker(Tracer):
    """
    Extends the Tracer class to invoke intervention functions.

    This class captures code blocks and compiles them into intervention functions
    that can be executed by the Interleaver.
    """

    def __init__(self, tracer: InterleavingTracer, *args, **kwargs):
        """
        Initialize an Invoker with a reference to the parent tracer.

        Args:
            tracer: The parent InterleavingTracer instance
            *args: Additional arguments to pass to the traced function
            **kwargs: Additional keyword arguments to pass to the traced function
        """
        
        if tracer is not None and tracer.model.interleaving:
            raise ValueError("Cannot invoke during an active model execution / interleaving.")
        
        self.tracer = tracer

        super().__init__(*args, **kwargs)

    def compile(self):
        """
        Compile the captured code block into an intervention function.

        The function is wrapped with try-catch logic to handle exceptions
        and signal completion to the mediator.

        Returns:
            A callable intervention function
        """
        self.info.source = [
            f"def __nnsight_tracer_{id(self)}__(__nnsight_mediator__, __nnsight_tracing_info__):\n",
            "    __nnsight_mediator__.pull()\n",
            *try_catch(
                self.info.source,
                exception_source=["__nnsight_mediator__.exception(exception)\n"],
                else_source=["__nnsight_mediator__.end()\n"],
            ),
        ]

        self.info.start_line -= 2

    def execute(self, fn: Callable):
        """
        Execute the compiled intervention function.

        Creates a new Mediator for the intervention function and adds it to the
        parent tracer's mediators list.

        Args:
            fn: The compiled intervention function
        """
        
        inputs, batch_group = self.tracer.batcher.batch(self.tracer.model, *self.args, **self.kwargs)

        self.inputs = inputs

        mediator = Mediator(fn, self.info, batch_group=batch_group)

        self.tracer.mediators.append(mediator)
